// +-------------------------------------------------------------------------+
// |                     IP4areas library vers. 0.4.50                       |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net>  2002-2007 |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// |     as published by Free Software Foundation (see the file LGPL.txt)    |
// |                                                                         |
// | Please visit http://www.croco.net/software/ip4areas to get a fresh copy |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#include "ip4range.hpp"

IP4Range::IP4Range(IP4IpAddress first, IP4IpAddress last)
{
    ip4addr_t i_f = first.HostByteOrderForm();
    ip4addr_t i_l = last.HostByteOrderForm();
    if(i_f>i_l) {
       Invalidate();
       return;
    }
    int len = i_l - i_f + 1;
    int min_possible_mask = 32;
    int mpm_range_len = 1;
    while(mpm_range_len < len) {
        min_possible_mask--;
        mpm_range_len <<= 1;
    }
    do {
        if(len == 1) {
            (*this) += IP4IpAddress(i_f);
            break;
        }
        while(mpm_range_len > len) {
            min_possible_mask++;
            mpm_range_len >>= 1;
        }
        // here, mpm_range_len must be no less than 1
        int mlen = min_possible_mask;
        while(mlen < 32 && (
                  ((i_f & ( ~((~(ip4addr_t)0)<<mlen))) != 0) ||
                  IP4Subnet(i_f,IP4Mask(mlen)).Last().HostByteOrderForm()>i_l
             ))
        {
            mlen++;
        }
        IP4Subnet sn(i_f,IP4Mask(mlen));
        (*this) += sn;
        i_f = sn.Last().HostByteOrderForm()+1;
        len = i_l - i_f + 1;
    } while(len > 0);
}
