// +-------------------------------------------------------------------------+
// |                     IP4areas library vers. 0.4.50                       |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net>  2002-2007 |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// |     as published by Free Software Foundation (see the file LGPL.txt)    |
// |                                                                         |
// | Please visit http://www.croco.net/software/ip4areas to get a fresh copy |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#include "ip4areas.hpp"
#include "ip4range.hpp"
#include "tests.hpp"


int main ()  {
    TestSection("IP4Areas");
    TestSubsection("IP4IpAddress");
    {
        IP4IpAddress *a1, *a2, *a3, *a4, *a5, *a6;
        a1 = new IP4IpAddress("1.2.3.4");
        a2 = new IP4IpAddress("2.3.4.5");
        a3 = new IP4IpAddress("1.2.3.4");
        a4 = new IP4IpAddress("invalid_address");
        a5 = new IP4IpAddress("666.777.888.999");
        a6 = new IP4IpAddress("255.255.255.255");
        TESTB("equality", (*a1 == *a3));
        TESTB("non-equality", (!(*a1 == *a2)));
        TESTB("valid1", ! a1->IsInvalid());
        TESTB("valid2", ! a2->IsInvalid()); 
        TESTB("valid3", ! a3->IsInvalid()); 
        TESTB("invalid1", a4->IsInvalid()); 
#ifndef __sun__
        TESTB("invalid2", a5->IsInvalid()); 
#endif
        TESTB("invalid3", a6->IsInvalid()); 
        TEST("textform", a1->TextForm(), "1.2.3.4");
        TESTB("less", *a1 < *a2);
        TESTB("no-less", !(*a2 < *a1));
        TESTB("eq-no-less", !(*a1 < *a3));
        delete a6;
        delete a5;
        delete a4;
        delete a3;
        delete a2;
        delete a1;
    }
    TestSubsection("IP4Mask");
    { 
        IP4Mask m1("255.255.255.240");
        IP4Mask m2("/27");
        IP4Mask m3(26);
        IP4Mask m4("255.255.255.255");
        IP4Mask m5("265.2.2.2");
        IP4Mask m6("invalid_mask");
        IP4Mask m7("255.254.255.240");
        TESTB("by_qdn", m1.Length() == 28);
        TESTB("by_suffix", m2.Length() == 27);
        TESTB("by_lenght", m3.Length() == 26);
        TESTB("valid1", ! m4.IsInvalid());
        TESTB("valid2", ! m1.IsInvalid());
        TESTB("invalid1", m5.IsInvalid());
        TESTB("invalid2", m6.IsInvalid());
        TESTB("invalid3", m7.IsInvalid());
        TESTB("extract", m1.BitMask() == 0xfffffff0);
        TEST("textform", m2.TextForm(), "/27");
        TEST("hostcount1", m1.HostCount(), 16);
        TEST("hostcount2", m2.HostCount(), 32);
        TEST("hostcount3", m3.HostCount(), 64);
        TEST("hostcount4", m4.HostCount(), 1);

        TEST("long_text_form", m2.LongTextForm(), "255.255.255.224");
    }
    TestSubsection("IP4Subnet");
    {
        IP4Subnet s1("194.87.251.18/24");
        IP4Subnet s2("194.87.251.18/32");
        IP4Subnet s3("194.87.251.18/32");
        IP4Subnet s4(IP4IpAddress("194.87.251.18"), IP4Mask(24));
        IP4Subnet s5("195.42.160.10/24");
        IP4Subnet s6("194.87.251.25/24");
        IP4Subnet s7("194.87.251.25/28");
        IP4Subnet s8("195.42.160.16/28");
        IP4Subnet s9("195.42.160.0/28");
        IP4Subnet s10("195.42.160.32/28");
        IP4Subnet s11("195.42.160.8/29");
        IP4Subnet s12("195.42.160.0/29");

        IP4Subnet s13("195.42.160.1"); // valid
        IP4Subnet s14("195.42.160.1/33"); // invalid
        IP4Subnet s15("195,42.160.1/33"); // invalid
        
        TESTB("addr/24", s1.GetPrefix() == IP4IpAddress("194.87.251.0"));
        TESTB("mask/24", s1.GetMask() == IP4Mask(24));
        TEST("subnet_hostcount", s1.HostCount(), 256);
        TESTB("equality", s1 == s4);
        TESTB("equality2", s2 == s3);
        TESTB("equality3", s1 == s6);
        TESTB("non-equality", ! (s1 == s3));
        TESTB("non-equality2", ! (s1 == s3));
        TEST("textform", s2.TextForm(), "194.87.251.18/32");
        TEST("textform2", s3.TextForm(), "194.87.251.18/32");
        TEST("textform2", s7.TextForm(), "194.87.251.16/28");
        TESTB("supernet", s5.IsSupernetOf(s8));
        TESTB("subnet", s8.IsSubnetOf(s5));
        TESTB("no-supernet", !s5.IsSupernetOf(s7));
        TESTB("no-subnet", !s5.IsSubnetOf(s8));
        TESTB("complementary", s8.IsComplementary(s9));
        TESTB("complementary2", s9.IsComplementary(s8));
        TESTB("non-complementary", !s8.IsComplementary(s10));
        TESTB("non-complementary2", !s8.IsComplementary(s11));
        TESTB("get-closest-supernet",
                  s11.GetClosestSupernet() == s9);
        TESTB("get-first-half", s9.GetFirstHalf() == s12);
        TESTB("get-second-half", s9.GetSecondHalf() == s11);

        TESTB("subnet_as_addr_valid", !s13.IsInvalid());
        TESTB("subnet_invalid2", s14.IsInvalid());
        TESTB("subnet_invalid3", s15.IsInvalid());

        IP4Subnet s16("192.168.5.19/28");
        TEST("subnet_first", s16.First().TextForm(), "192.168.5.16");
        TEST("subnet_last", s16.Last().TextForm(), "192.168.5.31");
    }
    TestSubsection("IP4Area");
    {
        IP4Area a1;
        TESTB("empty", a1.SubnetsCount() == 0);
        
        IP4Area a2;
        a2 += IP4Subnet("10.0.1.16/29");
        a2 += IP4Subnet("10.2.0.24/29");
        TEST("adding", a2.SubnetsCount(), 2);
        
        IP4Area a3;
        a3 += IP4Subnet("10.0.0.16/29");
        a3 += IP4Subnet("10.0.0.24/29");
        TESTB("complementary", a3.SubnetsCount() == 1);
        
        IP4Area a4;
        a4 += IP4Subnet("10.0.2.0/24");
        a4 += IP4Subnet("10.0.3.128/25");  // 128 - 255
        a4 += IP4Subnet("10.0.3.0/26");    // 0 - 63
        a4 += IP4Subnet("10.0.3.64/28");   // 64 - 79
        a4 += IP4Subnet("10.0.3.96/27");   // 96 - 127
        TESTB("non-aggregatable", a4.SubnetsCount() == 5);
        TEST("proper-order", a4.TextForm(),
           "10.0.2.0/24,10.0.3.0/26,10.0.3.64/28,10.0.3.96/27,10.0.3.128/25");
        TEST("first_in_area", a4.First().TextForm(), "10.0.2.0");
        TEST("last_in_area", a4.Last().TextForm(), "10.0.3.255");
        a4 += IP4Subnet("10.0.3.80/28");   // 80 - 95
        TEST("huge_aggregation", a4.SubnetsCount(), 1);
        
        IP4Area a5;
        a5.ReadIn("   10.0.0.0/24, 10.1.1.0/25,10.2.2.2/32");
        TEST("read-in",a5.TextForm(),"10.0.0.0/24,10.1.1.0/25,10.2.2.2/32");
        IP4Area::Iterator iter(a5);
        IP4Subnet s1;
        iter.GetNext(s1);
        TESTB("iteration1", s1 == IP4Subnet("10.0.0.0/24"));
        iter.GetNext(s1);
        TESTB("iteration2", s1 == IP4Subnet("10.1.1.0/25"));
        iter.GetNext(s1);
        TESTB("iteration3", s1 == IP4Subnet("10.2.2.2/32"));
        TESTB("iteration_end", !(iter.GetNext(s1)));
        TESTB("has_address", a5.HasAddress(IP4IpAddress("10.0.0.90")));
        TESTB("hasnt_address", !a5.HasAddress(IP4IpAddress("10.100.0.90")));
        TESTB("has_subnet", a5.HasSubnet(IP4Subnet("10.0.0.16/28")));
        TESTB("has_subnet2", a5.HasSubnet(IP4Subnet("10.1.1.0/25")));
        TESTB("has_subnet3", a5.HasSubnet(IP4Subnet("10.2.2.2/32")));
        TESTB("hasnt_subnet", !a5.HasSubnet(IP4Subnet("10.0.0.0/20")));
        TESTB("hasnt_subnet2", !a5.HasSubnet(IP4Subnet("10.2.2.2/31")));
        TESTB("hasnt_subnet3", !a5.HasSubnet(IP4Subnet("10.1.1.0/24")));

        IP4Area a6;
        a6.ReadIn("10.3.3.0/24, 10.0.0.0/16,10.2.2.0/24");
        a6 -= IP4Subnet("10.2.2.0/24");
        TEST("remove_exact_net", a6.SubnetsCount(), 2);
        a6 -= IP4Subnet("10.3.0.0/20");
        TEST("remove_supernet", a6.SubnetsCount(), 1);
        a6 -= IP4Subnet("10.0.17.0/24");
        TEST("remove_sub_subnet", a6.SubnetsCount(), 8);
#if 0 // no support for TESTS_lasttestfailed since the move to a new
      // test system
        if(TESTS_lasttestfailed)
            printf("%s\n", a6.TextForm());
#endif
        IP4Area a7;
        a7.ReadIn("10.10.10.0/29");
        a7 -= IP4Subnet("10.10.10.4/30");
        TEST("remove_second_half", a7.TextForm(),
                    "10.10.10.0/30");
        a7 -= IP4Subnet("10.10.10.0/31");
        TEST("remove_first_half", a7.TextForm(),
                    "10.10.10.2/31");

        IP4Area a8, a9;
        a8.ReadIn("10.0.0.0/16");
        a9.ReadIn("10.0.1.0/24, 10.0.128.0/18");
        a8 -= a9;
        TEST("remove_area_from_area", a8.SubnetsCount(), 8);

        IP4Area a10, a11;
        a10.ReadIn("10.0.10.0/24, 10.0.20.0/22");
        a11.ReadIn("10.0.128.0/17, 10.0.22.0/24");
        a10 += a11;
        TEST("add_area_to_area", a10.SubnetsCount(), 3);


        IP4Area a20;
        bool good = a20.ReadIn("some shit");
        TESTB("area_invalid", !good);
//        IP4Area a21;
//        a21.ReadIn("195.42.150.10/10, 195.7 20.25/10");
//        TESTB("area_invalid2", a21.IsInvalid());
//        IP4Area a22;
//        a22.ReadIn("10.20.30.40");
//        TESTB("area_invalid", a22.IsInvalid());

        
        IP4Area a30;
        a30.ReadIn("10.0.0.0/31, 10.12.14.16, 11.11.11.11");
        TEST("hostcount", a30.HostCount(), 4);
        IP4Area::HostIterator iter30(a30);
        IP4IpAddress ip30;
        iter30.GetNext(ip30);
        TESTB("hostiteration1", ip30 == IP4IpAddress("10.0.0.0"));
        iter30.GetNext(ip30);
        TESTB("hostiteration2", ip30 == IP4IpAddress("10.0.0.1"));
        iter30.GetNext(ip30);
        TESTB("hostiteration3", ip30 == IP4IpAddress("10.12.14.16"));
        iter30.GetNext(ip30);
        TESTB("hostiteration4", ip30 == IP4IpAddress("11.11.11.11"));
        TESTB("hostiteration_end", !(iter30.GetNext(ip30)));
        
        
    }
    TestSubsection("IP4Range");
    {
        {
        IP4Range r1(IP4IpAddress("10.0.0.1"),IP4IpAddress("10.0.0.255"));
        TEST("range", r1.TextForm(),
                             "10.0.0.1/32,10.0.0.2/31,10.0.0.4/30,"
                             "10.0.0.8/29,10.0.0.16/28,10.0.0.32/27,"
                             "10.0.0.64/26,10.0.0.128/25");
        TESTB("range_valid", !r1.IsInvalid());
        }
        {
        IP4Range r1(IP4IpAddress("10.0.0.7"),IP4IpAddress("10.0.0.37"));
        TEST("range2", r1.TextForm(),
                             "10.0.0.7/32,10.0.0.8/29,10.0.0.16/28,"
                             "10.0.0.32/30,10.0.0.36/31");
        TESTB("range2_valid", !r1.IsInvalid());
        }
        {
        IP4Range r1(IP4IpAddress("10.0.0.7"),IP4IpAddress("10.0.0.7"));
        TEST("range_same", r1.TextForm(), "10.0.0.7/32");
        TESTB("range_same_valid", !r1.IsInvalid());
        }
        {
        IP4Range r1(IP4IpAddress("10.0.0.7"),IP4IpAddress("10.0.0.6"));
        TESTB("range_invalid", r1.IsInvalid());
        }
                             
    }
    TestScore();
    return 0;
}
