// +-------------------------------------------------------------------------+
// |                     Script Plus Plus vers. 0.3.03                       |
// | Copyright (c) Andrey V. Stolyarov  <croco at croco dot net>  2003--2014 |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// |     as published by Free Software Foundation (see the file LGPL.txt)    |
// |                                                                         |
// | Please visit http://www.croco.net/software/scriptpp to get a fresh copy |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#include "scrvect.hpp"
#include "scrsubs.hpp"

ScriptSubstitution::~ScriptSubstitution()
{
    while(first) {
        item *p = first;
        first = p->next;
        delete p->var;
        delete p;
    }
}

ScriptSubstitution::VarId
ScriptSubstitution::AddVar(Var* v)
{
    item *p = new item;
    p->var = v;
    p->next = first;
    first = p;
    return first;
}

bool ScriptSubstitution::RemoveVar(ScriptSubstitution::VarId id)
{
    for(item **p = &first; *p; p = &(*p)->next) {
        if(*p == id) {
            item *tmp = *p;
            *p = tmp->next;
            delete tmp->var;
            delete tmp;
            return true;
        }
    }
    return false;
}

ScriptVariable ScriptSubstitution::Substitute(const ScriptVariable &src) const
{
    return DoSubstitution(src, first);
}

ScriptVariable ScriptSubstitution::
Substitute(const ScriptVariable &src,
           const ScriptVector &argv, int idx, int count) const
{
    item it;
    it.next = first;
    it.var = new ScriptSubstitutionArgv(&argv, false, idx, count);
    ScriptVariable res = DoSubstitution(src, &it);
    delete it.var;
    return res;
}

ScriptVariable ScriptSubstitution::
Substitute(const ScriptVariable &src, const Var *aux) const
{
    item it;
    it.next = first;
    it.var = aux;
    return DoSubstitution(src, &it);
}

ScriptVariable ScriptSubstitution::
DoSubstitution(const ScriptVariable &src, item *itemlist) const
{
    ScriptVariable res("");
    const char *p = src.c_str();
    while(*p) {
        if(*p == esc_char) {
            int idx = 0;
            p++;
            while(p[idx] && p[idx] != esc_char && p[idx] != '\n')
                idx++;
            if(!p[idx])
                break;
            if(p[idx] == '\n') {
                p += (idx + 1);
                continue;
            }
            if(idx == 0) {
                res += esc_char;
            } else {
                for(item *t = itemlist; t; t = t->next) {
                    ScriptVariable v = t->var->Sub(p, idx);
                    if(!v.IsInvalid()) {
                        res += v;
                        break;
                    }
                }
            }
            p += (idx + 1);
        } else {
            res += *p;
            p++;
        }
    }
    return res;
}

ScriptSubstitutionSingleName::
ScriptSubstitutionSingleName(const ScriptVariable &n)
    : name(n)
{}

ScriptVariable
ScriptSubstitutionSingleName::Sub(const char *nm, int len) const
{
    if(name.Length() != len)
        return ScriptVariableInv();
    for(int i = 0; i < len && nm[i]; i++)
        if(nm[i] != name[i])
            return ScriptVariableInv();
    return Text();
}


ScriptSubstitutionConst::
ScriptSubstitutionConst(const ScriptVariable &n, const ScriptVariable &v)
    : ScriptSubstitutionSingleName(n), val(v)
{}

ScriptVariable ScriptSubstitutionConst::Text() const
{
    return val;
}


ScriptSubstitutionScrVar::
ScriptSubstitutionScrVar(const ScriptVariable &n, ScriptVariable *pv)
    : ScriptSubstitutionSingleName(n), val(pv)
{}

ScriptVariable ScriptSubstitutionScrVar::Text() const
{
    return *val;
}



ScriptSubstitutionPrefixRequest::
ScriptSubstitutionPrefixRequest(const ScriptVariable &p, char d)
    : prefix(p), delim(d)
{
}

ScriptVariable
ScriptSubstitutionPrefixRequest::Sub(const char *name, int len) const
{
    if(len < prefix.Length()+1 || !ScriptVariable(name).HasPrefix(prefix))
        return ScriptVariableInv();

    ScriptVector params;
    int npar = 0;
    params[npar] = "";  /* to make sure it is valid */
    for(int i = prefix.Length()+1; i < len; i++)
        if(name[i] == delim) {
            npar++;
            params[npar] = ""; 
        } else {
            params[npar] += name[i]; 
        }
    return Handle(params);
}



ScriptSubstitutionArgv::
ScriptSubstitutionArgv(const ScriptVector *avec, bool copy, int aidx, int cnt)
    : vec_owned(copy), idx(aidx), count(cnt)
{
    if(copy) {
        vec = new ScriptVector(*avec, idx, count);
        idx = 0;
    } else {
        vec = avec;
    }
    if(count == -1 || count > (vec->Length() - idx))
        count = vec->Length() - idx;
}

ScriptSubstitutionArgv::~ScriptSubstitutionArgv()
{
    if(vec_owned)
        delete vec;
}

ScriptVariable ScriptSubstitutionArgv::Sub(const char *name, int len) const
{
    if(len == 1 && *name == '*') {
        ScriptVariable res("");
        for(int i = 1; i < count; i++) {
            if(i>1) res += ' ';
            res += (*vec)[i+idx];
        }
        return res;
    }
    int n = 0;    
    for(int i = 0; i < len; i++) {
        if(name[i] > '9' || name[i] < '0')
            return ScriptVariableInv();
        n *= 10;
        n += name[i] - '0';
    }
    if(n >= count)
        return ScriptVariable("");
    else
        return (*vec)[n+idx];
}

