//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <stdlib.h>

#include <scriptpp/scrvar.hpp>
#include <scriptpp/scrvect.hpp>
#include <sue/sue_sel.hpp>

#include "devices.hpp"
#include "flog.hpp"
#include "state.hpp"
#include "runfile.hpp"

#include "action.hpp"



static ScriptVector current_activity;

bool ready_for_security_now()
{
    return current_activity.Length() == 0;
}

static void add_movement(const ScriptVariable& mvid)
{
    bool was_empty = ready_for_security_now();
    for(int i=0; i<current_activity.Length(); i++) {
        if(current_activity[i] == mvid) return;  // already there
    }
    current_activity.AddItem(mvid);
    if(was_empty)
        state_process_event("!activity");
}

static void remove_movement(const ScriptVariable& mvid)
{
    bool was_empty = ready_for_security_now();
    for(int i=0; i<current_activity.Length(); ) {
        if(current_activity[i] == mvid)
            current_activity.Remove(i);
        else
            i++;
    }
    if(!was_empty && ready_for_security_now()) {
        state_process_event("!ready");
    }
}

static SUEEventSelector *the_selector = 0;
static bool should_restart = false;

class RestartTimeout : public SUETimeoutHandler {
public:
    RestartTimeout() {
        SetFromNow(2, 0);  // 2 seconds
        the_selector->RegisterTimeoutHandler(this);
    }
    ~RestartTimeout() {}
    virtual void TimeoutHandle() { the_selector->Break(); }
};


bool perform_action(const char *action)
{
    ScriptVector act(action);
    ScriptVariable n = act[0];
    n.Toupper();
    n.Trim();
    if(n == "@ACT+") {
        add_movement(act[1]);
    } else
    if(n == "@ACT-") {
        remove_movement(act[1]);
    } else
    if(n == "@LN+" || n == "@LN-") {
        return devices_output_action(act[1].c_str(), 
                                     (n[3]=='+') ? outact_on : outact_off);
    } else
    if(n == "@SET") {
        ScriptVariable var = act[1];
        act.Remove(0, 2);
        return runfile_set_variable(var.c_str(), act.Join(" ").c_str());    
    } else
    if(n == "@QUIT") {
        if(the_selector)
            new RestartTimeout;
        else
            exit(0);
    } else
    if(n == "@RESTART") {
        if(the_selector)
            new RestartTimeout;
        should_restart = true;
    } else {
        log_message("Unknown action [%s]", action);
        return false;
    }
    return true;
}

void action_memorize_selector(class SUEEventSelector *sel)
{
    the_selector = sel;
}

bool action_should_restart()
{
    return should_restart;
}
