//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <scriptpp/scrvar.hpp>
#include <scriptpp/scrvect.hpp>
#include <scriptpp/scrsubs.hpp>

#include "state.hpp"
#include "devices.hpp"
#include "onewire.hpp"
#include "config.hpp"
#include "flog.hpp"
#include "action.hpp"
#include "textsubs.hpp"
#include "commands.hpp"


// Builtin commands: &P &M &R &A &E

int builtin_args(char c)
{
    switch(c) {
    case 'P':
    case 'p':
    case 'M':
    case 'm':
        return 1;
    case 'R':
    case 'r':
    case 'A':
    case 'a':
    case 'E':
    case 'e':
    default:
        return 0;
    }
}

static void do_single_action(const ScriptVariable& a,
    const ScriptVector &args, int argidx, int argcount,
    ScriptVariable &response)
{
    log_message("** do_single_action(%s, [%s|%s], %d, %d)",
                a.c_str(), args[argidx].c_str(), args[argidx+1].c_str(),
                argidx, argcount);
    //ScriptSubstitution s;
    //s.AddVar(new ScriptSubstitutionArgv(&args, false, argidx, argcount+1));
    ScriptVariable act = FFSubstitution(a, args, argidx, argcount+1);
    if(act[0] != '@') {
        response += "Invalid action ``";
        response += act;
        response += "''\n";
        return;
    } 
    bool r = perform_action(act.c_str());
    if(!r) {
        response += "Action ``";
        response += act;
        response += "'' failed\n";
    }
}

static void do_action(const ScriptVariable& a,
    const ScriptVector &args, int argidx, int argcount,
    ScriptVariable &response)
{
    ScriptVector av(a, "\n", " \t\r");
    for(int i = 0; i < av.Length(); i++) {
        if(av[i] == "")
            continue;
        do_single_action(av[i], args, argidx, argcount, response);
    }
}

static void do_event(const ScriptVariable& a,
    const ScriptVector &args, int argidx, int argcount,
    ScriptVariable &response)
{
    //ScriptSubstitution s;
    //s.AddVar(new ScriptSubstitutionArgv(&args, false, argidx, argcount+1));
    ScriptVariable ev = FFSubstitution(a, args, argidx, argcount+1);
    ev.Trim();
    if(ev.Length() < 1) {
        response += "Empty event (?!); ignored\n";
        return;
    } 
    state_process_event(ev.c_str());
}

response_status process_command(const ScriptVariable &source_id,
                    const ScriptVariable &command, ScriptVariable &response)
{
    bool enabled = (source_id == "trusted");
    bool requested_answer = false;
#if 0
    bool need_mode_report = false;
#endif
    ScriptVector cmd(command);

    response = "";

    int i=0;
    while(i<cmd.Length()) {
        ScriptVariable content, cmd_resp;
        int argcount;
        bool unpriv;
        command_type ct = 
            config_decode_command(cmd[i].c_str(), content,
                                  argcount, cmd_resp, unpriv);
        if(enabled || unpriv) {
            log_message("Processing command %s(%d:%s)/%d...",
                        cmd[i].c_str(), ct, content.c_str(), argcount);
            switch(ct) {
            case cmdt_unknown:
                response += "Unknown cmd ";
                response += cmd[i];
                response += "\n";
                argcount = 0;
                break;
            case cmdt_builtin:
                argcount = builtin_args(content[1]);
                switch(content[1]) {
                case 'p':
                case 'm':
                    if(i+1 < cmd.Length())
                        enabled = config_authentication(
                            content[1] == 'p' ? source_id : ScriptVariable(""),
                            cmd[i+1]);
                    break;
                case 'r':
                    requested_answer = true;
                    if(cmd_resp.IsInvalid()) {
                        response += devices_report();
                        response += onewire_report();
#if 0
                        need_mode_report = true;
#endif
                        requested_answer = true;
                    }
                    break;         
                case 'a':
                    requested_answer = true;
                    break;
                case 'e':
                    i = cmd.Length(); // so the loop will quit
                }
                break;
            case cmdt_action:
                if(argcount == -1) argcount = cmd.Length() - i;
                do_action(content, cmd, i, argcount, response);
                break;
            case cmdt_event:
                if(argcount == -1) argcount = cmd.Length() - i;
                do_event(content, cmd, i, argcount, response);
#if 0
                need_mode_report = true;
#endif
                break;
            case cmdt_response: // no action, just make the response
                                // which is done later anyway
                argcount = 0;
                break;
            default:  // impossible :-)
                argcount = 0;
            }
            if(cmd_resp.IsValid()) {
                ScriptVariable rr =
                    FFSubstitution(cmd_resp, cmd, i, argcount+1);
                response += rr;
            }
        } else {
            log_message("Rejecting command %s(%d:%s)/%d...",
                        cmd[i].c_str(), ct, content.c_str(), argcount);
            return response_none;
        }
        i += (argcount >= 0 ? argcount : 0)  + 1;
    }
#if 0
    if(need_mode_report) {
        response += " Mode: ";
        response += get_current_state();
        response += "\n";
    }
#endif
    if(!enabled || response == "")
        return response_none;
    return requested_answer ? response_important : response_info;
}








