//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <inifile/inifile.hpp>
#include <scriptpp/scrvar.hpp>
#include <scriptpp/scrvect.hpp>


#include "flog.hpp"
#include "textsubs.hpp"
#include "config.hpp"


static IniFileParser *config = 0;


#define DEFAULT_SYSNAME "unnamed_ffazenda"
#define DEFAULT_LOGFILE "ffazenda.log"
#define DEFAULT_RUNFILE "ffazenda.state"
#define DEFAULT_DEFSTATE "#nostate#"
#define DEFAULT_WATCHDOG 30

static ScriptVariable config_file_name("");

bool init_config(const char *conf_name)
{
    config_file_name = conf_name;
    if(config)
        delete config;
    config = new IniFileParser();
    if(!config->Load(conf_name)) {
        log_message("%s:%d: %s", conf_name,
                    config->GetLastErrorLine(),
                    config->GetLastErrorDescription());
        log_message("%s: error loading configuration file", conf_name);
        return false;
    }
    return true;
}

void config_report()
{
    log_message("Configuration: %s", config_file_name.c_str());
    log_message("System name: %s",
                config->GetTextParameter("general", 0, "name", "*NONAME*"));
}

ScriptVariable config_get_system_name()
{
    if(!config)
        return DEFAULT_SYSNAME;
    return config->GetTextParameter("general", 0, "name", DEFAULT_SYSNAME);
}

ScriptVariable config_get_log_name()
{
    if(!config)
        return DEFAULT_LOGFILE;
    return config->GetTextParameter("general", 0, "log", DEFAULT_LOGFILE);
}

ScriptVariable config_get_runfile_name()
{
    if(!config)
        return DEFAULT_RUNFILE;
    return config->GetTextParameter("general", 0, "state", DEFAULT_RUNFILE);
}

ScriptVariable config_get_defaultstate()
{
    if(!config)
        return DEFAULT_DEFSTATE;
    return config->GetTextParameter("general", 0, "defaultstate",
                                    DEFAULT_DEFSTATE);
}

int config_get_watchdog_interval()
{
    if(!config)
        return DEFAULT_WATCHDOG;
    return config->GetIntegerParameter("general", 0, "watchdog",
                                       DEFAULT_WATCHDOG);
}

bool config_runvar_exists(const char *name)
{
    if(!config)
        return false;
    const char *ven = config->GetTextParameter("variable", name, "enable", 0);
    if(!ven)
        return false;
    return ScriptVariable(ven).Trim().Toupper() == "YES";
}

ScriptVariable config_runvar_default(const char *name)
{
    if(!config_runvar_exists(name))
        return ScriptVariableInv();
    return config->GetTextParameter("variable", name, "default", "");
}

#if 0
int config_security_timeout_seconds()
{
    return config->GetIntegerParameter("general", 0, "security_timeout", 30);
}
int config_full_alert_timeout_seconds(bool paranoid)
{
    return config->GetModifiedIntegerParameter("general", 0,
               "full_alert_timeout", paranoid ? "paranoid" : "relaxed", 30);
}
#endif

int config_get_devices_count()
{
    return config->GetSectionCount("device");
}

bool config_get_device_config(int idx, device_config &conf)
{
    const char *name = config->GetSectionName("device", idx);
    if(!name || !*name) return false;
    long id;
    if(!ScriptVariable(name).GetLong(id, 16)) {
        log_message("Illegal device id ``%s'' in conf. file, ignoring",name);
        return false;
    }
    conf.id = id;
    for(int i=0; i<12; i++) {
        ScriptNumber n(i);
        const char *par;

        par = config->GetModifiedTextParameter(
            "device", name, "output", n.c_str(), 0
        );
        if(par)
            conf.output[i] = ScriptVariable(par).Trim();
        else
            conf.output[i].Invalidate();
        par = config->GetModifiedTextParameter(
            "device", name, "output_sync", n.c_str(), 0
        );
        conf.output_sync[i] =
            par && ScriptVariable(par).Trim().Tolower() == "yes";

        if(i>=8)
            continue; // there can be 12 outs, but no more than 8 ins

        par = config->GetModifiedTextParameter(
            "device", name, "input", n.c_str(), 0
        );
        if(par)
            conf.input_name[i] = ScriptVariable(par).Trim();
        else
            conf.input_name[i].Invalidate();
        par = config->GetModifiedTextParameter(
            "device", name, "input_on", n.c_str(), 0
        );
        if(par)
            conf.input_on[i] = ScriptVariable(par).Trim();
        else
            conf.input_on[i].Invalidate();
        par = config->GetModifiedTextParameter(
            "device", name, "input_off", n.c_str(), 0
        );
        if(par)
            conf.input_off[i] = ScriptVariable(par).Trim();
        else
            conf.input_off[i].Invalidate();
    }
    const char *p1w = config->GetTextParameter("device", name, "onewire", 0);
    conf.poll_onewire = p1w && (ScriptVariable(p1w).Tolower() == "yes");
    return true;
}

int config_get_tsens_count()
{
    return config->GetSectionCount("temp_sensor");
}

static int hexdig2num(char c)
{
    if(c >= '0' && c <= '9') return c - '0';    
    if(c >= 'A' && c <= 'F') return c - 'A' + 10;    
    if(c >= 'a' && c <= 'f') return c - 'a' + 10;    
    return -1;
}

static void
add_watermark_to_tsconf(ScriptVariable &s, temp_sensor_config &conf, int idx)
{
    s.Trim();
    if(s == "")
        return; // it's okay, simply ignore the empty line
    ScriptVector v(s, ":", " \n\t\r");
        // it's smth. like this:
        // 5:20:climate freezes:climate warms:3:starting cold:starting hot
    if(v.Length() < 4) {
        log_message("Invalid watermark: ``%s'', ignoring", s.c_str());
        return;
    }
    long lm, hm, dly;
    if(!v[0].GetLong(lm, 10)) {
        log_message("Invalid low temp. for: ``%s'', ignoring", s.c_str());
        return;
    }
    if(!v[1].GetLong(hm, 10)) {
        log_message("Invalid high temp. for: ``%s'', ignoring", s.c_str());
        return;
    }
    if(v.Length() >=5) {
        if(!v[4].GetLong(dly, 10)) {
            log_message("Invalid delay: ``%s'', ignoring", s.c_str());
            return;
        }
    } else {
        log_message("No delay: ``%s'', defaulting to 2", s.c_str());
        dly = 2;
    }
    temp_sensor_config::watermark *w = new temp_sensor_config::watermark;
    w->index = idx;
    w->low_mark = lm;
    w->high_mark = hm;
    w->transit_delay = dly;
    w->low_event = v[2];
    w->high_event = v[3];
    if(v.Length() >= 6)
        w->low_init_event = v[5];
    if(v.Length() >= 7)
        w->high_init_event = v[6];
    w->next = conf.first;
    conf.first = w;
}

bool config_get_tsens_config(int idx, temp_sensor_config &conf)
{
    int i;
    const char *name = config->GetSectionName("temp_sensor", idx);
    if(!name || !*name) return false;
    conf.name = name;

    const char *id = config->GetTextParameter("temp_sensor", name, "id", 0);
    if(!id) {
        log_message("No id for temp. sensor ``%s'' in conf. file", name);
        return false;
    }
    const char *p = id;
    for(i=0; i<8; i++) {
        if(!*p || !*(p+1)) {
            log_message("Too short temp. sensor id ``%s'' in conf. file", id);
            return false;
        }
        int d1 = hexdig2num(*p++);
        int d2 = hexdig2num(*p++);
        if(d1 == -1 || d2 == -1) {
            log_message("Illegal temp. sensor id ``%s'' in conf. file", id);
            return false;
        }
        conf.id[7-i] = (d1 << 4) | d2;
    }

    const char *rep =
        config->GetTextParameter("temp_sensor", name, "report", "no");
    conf.report = (rep && (ScriptVariable(rep).Toupper() == "YES"));

    const char *wm =
        config->GetTextParameter("temp_sensor", name, "watermarks", 0);
    if(!wm) return true; // it's okay 

    ScriptVariable wm_sub = FFSubstitution(wm);

    ScriptVector vec(wm_sub, "\n", " \t\r");
        // break down to lines, trimming all leading and trailing spaces
    if(vec[0] == "")
        vec.Remove(0, 1);
    for(i=0; i<vec.Length(); i++)
        add_watermark_to_tsconf(vec[i], conf, i+1);

    return true;
}

const char *config_get_console_greeting(bool trusted)
{
    return config->GetModifiedTextParameter("console", 0, "greeting",
                         trusted ? "trusted" : "untrusted", "");
}
const char *config_get_console_ip()
{
    return config->GetTextParameter("console", 0, "ip", "127.0.0.1");
}
int config_get_console_port()
{
    return config->GetIntegerParameter("console", 0, "port", 7280);
}
int config_get_console_timeout_seconds()
{
    return config->GetIntegerParameter("console", 0, "timeout", 60);
}
const char *config_get_console_trusted_networks()
{
    return config->GetTextParameter("console", 0, "trusted", "");
}

const char *config_get_webserver_ip()
{
    return config->GetTextParameter("webserver", 0, "ip", "127.0.0.1");
}
int config_get_webserver_port()
{
    return config->GetIntegerParameter("webserver", 0, "port", 7281);
}
const char *config_get_webserver_realm()
{
    return config->GetTextParameter("webserver", 0, "realm", "127.0.0.1");
}
int config_get_webserver_timeout_seconds()
{
    return config->GetIntegerParameter("webserver", 0, "timeout", 600);
}
const char *config_get_webserver_content(const char *content_id)
{
    return config->GetModifiedTextParameter("webserver", 0,
                                            "html", content_id, 0);
}
const char *config_get_webserver_p404()
{
    return config->GetTextParameter("webserver", 0, "p404", "Not found");
}
ScriptVariable config_get_webserver_password_by_user(const char *user)
{
    const char *r = 
        config->GetTextParameter("webserver", 0, "users", "");
    //log_message("web users: %s", r);
    ScriptTokenVector v(r, ",\n", " \r\t");
    for(int i = 0; i < v.Length(); i++) {
        ScriptTokenVector uu(v[i], ":", " \r\n\t");
        if(uu[0] == user)
            return uu[1];
    }
    return ScriptVariableInv();
}

bool config_get_webpage_properties(const char *name, webpage_properties *p)
{
    int alias_count = 0;
    const char *alias;
    do {
        if(++alias_count > 20) {
            log_message("config. err: too many redirections for %s", name);
        }
        alias = config->GetTextParameter("webpage", name, "alias_for", 0);
        if(alias)
            name = alias;
    } while(alias);
    const char *type = config->GetTextParameter("webpage", name, "type", 0);
    if(!type)
        return false;
    p->type = type;
    const char *mime = config->GetTextParameter("webpage",name,"mimetype",0);
    p->mime_type = mime ? mime : "text/html";
    const char *pub = config->GetTextParameter("webpage", name, "public", 0);
    p->is_public = false;
    if(pub) {
        ScriptVariable pubs(pub);
        pubs.Tolower();
        if(pubs == "yes")
            p->is_public = true;
    }
    if(p->type == "page" || p->type == "command") {
        p->path.Invalidate();
        const char *cont =
            config->GetTextParameter("webpage", name, "content", 0);
        if(!cont) {
            log_message("conf. error: no content for %s", name);
            return false;
        }
        p->content = cont;
        if(p->type == "command") {
            const char *resp =
                config->GetTextParameter("webpage", name, "response", 0);
            p->response = resp ? resp : p->content;
        } else {
            p->response.Invalidate();
        }
    } else
    if(p->type == "file") {
        p->content.Invalidate();
        p->response.Invalidate();
        const char *path =
            config->GetTextParameter("webpage", name, "path", 0);
        if(!path) {
            log_message("conf. error: no path for %s", name);
            return false;
        }
        p->path = path;
    } else {
        log_message("conf. error: page type ``%s'' not supported", type);
        return false;
    }
    return true;
}

const char *config_get_sms_incoming_dir()
{
    return config->GetTextParameter("sms", 0, "incoming", 0);
}
const char *config_get_sms_outgoing_dir()
{
    return config->GetTextParameter("sms", 0, "outgoing", "/var/sms/out");
}
const char *config_get_sms_composing_dir()
{
    return config->GetTextParameter("sms", 0, "composing","/var/ffazenda/sms");
}
const char *config_get_sms_proceeded_dir()
{
    return config->GetTextParameter("sms", 0,
                                      "proceeded","/var/ffazenda/sms_done");
}
const char *config_get_sms_failed_dir()
{
    return config->GetTextParameter("sms", 0, "failed", 0);
}

bool config_authentication(const ScriptVariable &source,
                           const ScriptVariable &password)
{
    if(source == "") {  // check the master password
        const char *p =
            config->GetTextParameter("general", 0, "password", 0);
        return p && (password == p);
    } else { // check the auth. scheme and the password for the source
        const char *scm = 
            config->GetTextParameter("phone", source.c_str(), "auth", 0);
        if(ScriptVariable(scm) == "trusted") return true;
        const char *p = 
            config->GetTextParameter("phone", source.c_str(), "password", 0);
        return p && (password == p);
    }
}

bool config_get_state_properties(const char *state,
                                 int &lmon, int &lmoff, sound_modes &snd,
                                 ScriptVariable &notific,
                                 int &trans_tm, ScriptVariable &trans_st,
                                 ScriptVariable &trans_event,
                                 ScriptVariable &unready_st)
{
    ScriptVariable lamp = config->GetTextParameter("state", state, "lamp", "");
    if(lamp == "") {
        lmon = lmoff = -1;
    } else
    if(lamp == "on") {
        lmon = 1;
        lmoff = 0;
    } else
    if(lamp == "off") {
        lmon = 0;
        lmoff = 0;
    } else {
        ScriptVector lv(lamp, ":", " \t\n");
        long a, b;
        if(lv.Length() == 2 && lv[0].GetLong(a) && lv[1].GetLong(b)) {
            lmon = a;
            lmoff = b;
        } else {
            log_message("Incorrect lamp timing for state ``%s''", state);
            return false;
        }
    }
    ScriptVariable sound = config->GetTextParameter("state",state,"sound","");
    sound.Trim();
    sound.Tolower();
    if(sound == "")
        snd = snd_mode_unchanged;
    else if(sound == "off")
        snd = snd_mode_off;
    else if(sound == "on")
        snd = snd_mode_on;
    else if(sound == "withlamp")
        snd = snd_mode_withlamp;
    else if(sound == "against")
        snd = snd_mode_against;
    else {
        log_message("Unknown sound mode ``%s'' for state ``%s''",
                    sound.c_str(), state);
        return false;
    }

    const char *ntf =
        config->GetTextParameter("state", state, "notification", "");
    if(ntf && *ntf)
        notific = ntf;
    else
        notific.Invalidate();

    const char *tr = config->GetTextParameter("state", state, "transit", 0);
    if(tr) {
        ScriptVector trr(tr, ":", " \t\n");
        long trt;
        if(trr.Length() < 2 || !trr[0].GetLong(trt)) {
            log_message("Incorrect transit spec for state ``%s''", state);
            return false;
        }
	trans_tm = trt;
        trans_st = trr[1];
        if(trr.Length() >= 3)
            trans_event = trr[2];
        else
            trans_event.Invalidate();
    } else {
        trans_tm = -1;
        trans_st.Invalidate();
        trans_event.Invalidate();
    }
    const char *mt = config->GetTextParameter("state", state, "unready", 0);
    if(mt) {
        unready_st = mt;
    } else {
        unready_st.Invalidate();
    }

    return true;
}

const char *config_get_event_fsm(const char *event)
{
    return config->GetTextParameter("event", event, "fsm", "");
}

const char *config_get_event_transit(const char *event, const char *selstate)
{
    return config->GetModifiedTextParameter("event", event,
                                            "transit", selstate, 0);
}
const char *config_get_event_action(const char *event, const char *selstate)
{
    return config->GetModifiedTextParameter("event", event,
                                            "action", selstate, 0);
}
const char *
config_get_event_notification(const char *event, const char *selstate)
{
    return config->GetModifiedTextParameter("event", event,
                                            "notification", selstate, 0);
}
const char *
config_get_event_exec(const char *event, const char *selstate)
{
    return config->GetModifiedTextParameter("event", event,
                                            "exec", selstate, 0);
}

const char *
config_get_event_subevents(const char *event, const char *selstate)
{
    return config->GetModifiedTextParameter("event", event,
                                            "subevents", selstate, 0);
}

const char *config_get_notification_phones(const char *notific)
{
    return config->GetTextParameter("notification", notific, "phones", 0);
}

const char *config_get_notification_message(const char *notific)
{
    return config->GetTextParameter("notification", notific, "message", 0);
}

command_type config_decode_command(const char *cmd, ScriptVariable &content,
    int &argcount, ScriptVariable &response, bool &unpriv)
{
    bool resolved = false;
    unpriv = false;
    ScriptVariable cmds(cmd);
    response.Invalidate();  // the first found in the chain will be active
    do {
        cmds.Tolower();
        if(cmds[0] == '&') {
            content = cmds;
            argcount = -1;
            unpriv = true;
            return cmdt_builtin;
        }
        if(response.IsInvalid()) {  // may be there is a response here?
            const char *resp =
                config->GetTextParameter("command",cmds.c_str(),"response",0);
            if(resp) {
                response = resp;
            }
        }
        if(!unpriv) { // may be permission here?
            const char *up =
                config->GetTextParameter("command", cmds.c_str(), "unpriv", 0);
            if(up) {
                ScriptVariable ups(up);
                ups.Tolower();
                if(ups == "yes")
                    unpriv = true;
            }
        }
        const char *c =
            config->GetTextParameter("command", cmds.c_str(), "alias_for", 0);
        if(!c)
            resolved = true;
        else
            cmds = c;
    } while(!resolved);
    cmd = cmds.c_str();
    argcount = config->GetIntegerParameter("command", cmd, "args", -1);
    const char *s = config->GetTextParameter("command", cmd, "action", 0);
    if(s) {
        content = s;
        return cmdt_action;
    }
    s = config->GetTextParameter("command", cmd, "event", 0);
    if(s) {
        content = s;
        return cmdt_event;
    }
    if(response.IsValid())
        return cmdt_response;
    return cmdt_unknown;
}
