//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#ifndef CONFIG_HPP_SENTRY
#define CONFIG_HPP_SENTRY

#include <scriptpp/scrvar.hpp>

bool init_config(const char *conf_name);
void config_report();

ScriptVariable config_get_system_name();
ScriptVariable config_get_log_name();
ScriptVariable config_get_runfile_name();
ScriptVariable config_get_defaultstate();

int config_get_watchdog_interval();

bool config_authentication(const ScriptVariable &source,
                           const ScriptVariable &password);



/////////////////////////////////////////////////
// runtime variable configs

bool config_runvar_exists(const char *name);
ScriptVariable config_runvar_default(const char *name);

/////////////////////////////////////////////////
// device configs

struct device_config {
    int id;
    bool poll_onewire;
    ScriptVariableInv input_name[8];
    ScriptVariableInv input_on[8];
    ScriptVariableInv input_off[8];
    ScriptVariableInv output[12];
    bool output_sync[12];
};

int config_get_devices_count();
bool config_get_device_config(int idx, device_config &conf);


/////////////////////////////////////////////////
// temperature sensors 

/* Please note that temperatures are ALWAYS in integer degrees of Celsius.
   They are converted from the Dallas/Maxim notation immediately upon
   retrieval.
 */

struct temp_sensor_config {
    enum wm_state { wms_undef, wms_low, wms_middle, wms_high };
        // should wms_* change one day, their names in onewire.cpp MUST
        // be changed accordingly 
    struct watermark {
            // Actually, state, next_state and transit_counter have
            // nothing to do with the configuration, but it is too
            // ugly to derive something new in the derived class
        wm_state state, next_state;
        int transit_counter;
        int index, low_mark, high_mark, transit_delay;
        ScriptVariable low_event, high_event,
            low_init_event, high_init_event;
        watermark *next;
    
        watermark() :
            state(wms_undef), next_state(wms_undef), transit_counter(0),
            index(0), low_mark(-1000), high_mark(1000), transit_delay(2),
            low_event(""), high_event(""),
            low_init_event(""), high_init_event(""), next(0)
        {}
        ~watermark() { if(next) delete next; }
    };

    ScriptVariable name;
    union {
        unsigned char id[8];
        unsigned long long id_numerical;
    };
    bool report;
    watermark *first;

    temp_sensor_config() : first(0) {}
    ~temp_sensor_config() { if(first) delete first; }
};

int config_get_tsens_count();
bool config_get_tsens_config(int idx, temp_sensor_config &conf);


//////////////////////////////////////
// misc

const char *config_get_console_greeting(bool trusted);
const char *config_get_console_ip();
int config_get_console_port();
int config_get_console_timeout_seconds();
const char *config_get_console_trusted_networks();

const char *config_get_webserver_ip();
int config_get_webserver_port();
int config_get_webserver_timeout_seconds();
const char *config_get_webserver_realm();
const char *config_get_webserver_content(const char *content_id);
const char *config_get_webserver_p404();
ScriptVariable config_get_webserver_password_by_user(const char *user);

struct webpage_properties {
    ScriptVariable type;
    bool is_public;
    ScriptVariable content;
    ScriptVariable response;
    ScriptVariable path;
    ScriptVariable mime_type;
};

bool config_get_webpage_properties(const char *name, webpage_properties *p);

const char *config_get_sms_incoming_dir();
const char *config_get_sms_composing_dir();
const char *config_get_sms_outgoing_dir();
const char *config_get_sms_proceeded_dir();
const char *config_get_sms_failed_dir();

enum sound_modes {
    snd_mode_unchanged,
    snd_mode_off, snd_mode_on, snd_mode_withlamp, snd_mode_against
};

bool config_get_state_properties(const char *state,
                                 int &lmon, int &lmoff, sound_modes &snd,
                                 ScriptVariable &notific,
                                 int &trans_tm, ScriptVariable &trans_st,
                                 ScriptVariable &trans_event,
                                 ScriptVariable &unready_st);

const char *config_get_event_fsm(const char *event);
const char *config_get_event_transit(const char *event, const char *selstate);
const char *config_get_event_action(const char *event, const char *selstate);
const char *config_get_event_notification(const char *event, const char *ss);
const char *config_get_event_exec(const char *event, const char *selstate);
const char *config_get_event_subevents(const char *event, const char *ss);


const char *config_get_notification_phones(const char *notific);
const char *config_get_notification_message(const char *notific);

enum command_type {
    cmdt_unknown, cmdt_builtin, cmdt_action, cmdt_event, cmdt_response
};

command_type
config_decode_command(const char *cmd, ScriptVariable &content, int &argcount,
                      ScriptVariable &response, bool &unpriv);


#endif
