//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <sue/sue_sel.hpp>
#include <sue/sue_tcps.hpp>

#include <ip4areas/ip4areas.hpp>


#include "config.hpp"
#include "flog.hpp"
#include "commands.hpp"
#include "console.hpp"

class ConsoleServerSession : public SUETcpServerSession {
    bool trusted;
public:
    ConsoleServerSession(int a_fd, int a_timeout, 
                      SUEEventSelector *a_selector,
                      class ConsoleServer *a_server);

    ~ConsoleServerSession();
    virtual void HandleNewInput();
    virtual void TcpServerSessionShutdownHook(); 
    virtual void HandleSessionTimeout();
};

class ConsoleServer : public SUETcpServer {
    int timeout;
public:
    ConsoleServer(const char *ip, int a_port, int a_timeout);
    ~ConsoleServer();

    virtual SUETcpServerSession* SpawnSession(int newsessionfd);
};


ConsoleServerSession::ConsoleServerSession(int a_fd, int a_timeout, 
	      SUEEventSelector *a_selector,
	      ConsoleServer *a_server)
    : SUETcpServerSession(a_fd, a_timeout, a_selector, a_server, ""),
      trusted(false)
{ 
    inputresetstimeout = true;
    outputresetstimeout = false;

    unsigned long addr = a_server->GetIpOfLastAccepted();

    IP4Area trusted_area;
    trusted_area.ReadIn(config_get_console_trusted_networks());
    if(trusted_area.HasAddress(addr)) trusted = true;

    log_message("Console: connection from %x:%d -- %s",
                addr, a_server->GetPortOfLastAccepted(),
                trusted ? "trusted" : "untrusted");

    outputbuffer.AddString(config_get_console_greeting(trusted));
}

ConsoleServerSession::~ConsoleServerSession() 
{
    log_message("Console session closed (destructor)");
}

void ConsoleServerSession::HandleNewInput() 
{   
    SUEBuffer ln;
    while(inputbuffer.ReadLine(ln)) {
        log_message("Console: handling [%s]", ln.GetBuffer());
        if(ln.Length() < 1 || ln[0] == 0) {
            GracefulShutdown();
        } else {
            ScriptVariable response;
            response_status s = process_command(trusted ? "trusted" : "",
                                                ln.GetBuffer(), response);
            if(s != response_none)
            {
                if(s == response_important)
                    outputbuffer.AddString("[!] ");
                outputbuffer.AddString(response.c_str());
                outputbuffer.AddString("\n");
            }
        }
    }
}

void ConsoleServerSession::HandleSessionTimeout() 
{
    outputbuffer.AddString("Timed out, sorry...\n");
    GracefulShutdown();
}

void ConsoleServerSession::TcpServerSessionShutdownHook() 
{
    // actually, nothing to do here...
    // the object will be deleted by the server
    log_message("Console: shutting down the session");
}

ConsoleServer::ConsoleServer(const char *ip, int a_port, int a_timeout)
    : SUETcpServer(ip, a_port)
{
    timeout = a_timeout;
}

ConsoleServer::~ConsoleServer()
{}

SUETcpServerSession* ConsoleServer::SpawnSession(int newsessionfd)
{
    return new ConsoleServerSession(newsessionfd, timeout, selector, this);
}

static ConsoleServer *server = 0;

void console_init(class SUEEventSelector *selector)
{
    if(server) { delete server; server = 0; }
    const char *ip = config_get_console_ip();
    int port = config_get_console_port();
    int tm = config_get_console_timeout_seconds();
    server = new ConsoleServer(ip, port, tm);
    server->Up(selector);
}
