//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <sue/sue_sel.hpp>
#include "unc0xx.h"
#include "unc_data.h"

#include "config.hpp"
#include "runfile.hpp"
#include "flog.hpp"
#include "state.hpp"
#include "onewire.hpp"

#include "devices.hpp"



struct device_item {
    unc0xx_devhdl *dev_hdl;
    long version;
    device_config conf;
    int min_cnt, max_cnt;
    bool poll_inputs;
    int last_in_state[8];
    int last_state;
//    int out_state_as_set[12];
    bool output_sync_needed;
    device_item *next;
    device_item()
        : dev_hdl(0), conf(), min_cnt(-1), max_cnt(-1),
          poll_inputs(false), last_state(-1), output_sync_needed(false) {}
};

device_item *devices = 0;

static void drop_devices()
{
    while(devices) {
        device_item *t = devices;
        devices = devices->next;
        if(t->dev_hdl)
            unc0xx_close(t->dev_hdl);
        delete t;
    }
}

static void connect_to_device(device_item *t, bool failure_diags)
{
    unc0xx_usb_rescan();
    unc0xx_scan_item *dev = unc0xx_scan(t->conf.id);
    if(dev) {
        if(dev->dev_hdl) {
            log_message("UNC0xx %x is [%x:%x:%x], vers. %d, good",
                    t->conf.id,
                    dev->sernum[0], dev->sernum[1], dev->sernum[2],
                    dev->version);
            t->dev_hdl = dev->dev_hdl;
            t->version = dev->version;
            unc0xx_get_counter_range(t->dev_hdl,
                                     &t->min_cnt, &t->max_cnt);
            int rc = unc0xx_zero_query_counters(t->dev_hdl,
                                       t->last_in_state, 8);
            if(rc == -1) {
                // no, failed
                t->dev_hdl = 0;  // remains unconnected
                   // connection will be closed by unc0xx_destroy_list
                if(failure_diags)
                    log_message("UNC0xx device %x counters failure",
                        t->conf.id);
            } else {
                for(int i=0; i<8; i++)
                    t->last_in_state[i] &= 1;
                dev->dev_hdl = 0; // well we take it over
            }
        } else {
            if(failure_diags)
                log_message("UNC0xx %x has incompatible version! "
                            "Please upgrade the firmware");
        }
        unc0xx_destroy_list(dev);
    } else {
        if(failure_diags)
            log_message("UNC0xx device %x not found, check connections",
                        t->conf.id);
    }
}

static void device_lost(device_item *t)
{
    if(t->dev_hdl)
        unc0xx_close(t->dev_hdl);
    t->dev_hdl = 0;
    log_message("UNC0xx %04x is LOST!", t->conf.id);
}

void devices_init(SUEEventSelector *selector)
{
    static bool usb_init = false;
    if(!usb_init) {
        unc0xx_usb_init();
        usb_init = true;
    }
    drop_devices();
    unc0xx_usb_rescan();
    int n = config_get_devices_count();
    log_message("Total %d device configuration entries found", n);
    for(int i=0; i<n; i++) {
        device_config conf;
        if(config_get_device_config(i, conf)) {
            device_item *t = new device_item;
            t->conf = conf;
            t->next = devices;
            devices = t;
            for(int i=0; i<12; i++) {
                if(t->conf.output[i].IsValid()) {
                    log_message("Device ``%x'' output %d: %s %s",
                                t->conf.id, i, t->conf.output[i].c_str(),
                                t->conf.output_sync[i] ? "SYNC" : "unsync");
                    if(t->conf.output_sync[i]) {
                        t->output_sync_needed = true;
#if 0   // seems that we don't really need to sync anything here
                        int runst =
                            runfile_get_output(t->conf.output[i].c_str());
                        switch(runst) {
                            case runfile_output_on:
                                output_state_as_set[i] = 1;
                                break;
                            case runfile_output_off:
                                output_state_as_set[i] = -1;
                                break;
                            case runfile_output_unset:
                            default:
                                output_state_as_set[i] = 0;
                                break;
                        } else {
                            output_state_as_set[i] = 0;
                        }
#endif
                    }
                }

                if(i>=8)
                    continue;

                if(t->conf.input_on[i].IsValid()) {
                    log_message("Device ``%x'' input %d on conn: %s",
                                t->conf.id, i, t->conf.input_on[i].c_str());
                    t->poll_inputs = true;
                }
                if(t->conf.input_off[i].IsValid()) {
                    log_message("Device ``%x'' input %d on discn: %s",
                                t->conf.id, i, t->conf.input_off[i].c_str());
                    t->poll_inputs = true;
                }

            }
            connect_to_device(t, true);
        }
    }
}

static void do_perform_action(device_item *p, int i, int act)
{
    int action = ((i/4) << 8) | (act << ((i%4) * 2));
    unc0xx_perform_action(p->dev_hdl, action);
}

bool devices_output_action(const char *channel_id, output_actions act)
{
    for(device_item *p = devices; p; p = p->next)
        for(int i=0; i<12; i++)
            if(p->conf.output[i] == channel_id) {
                if(p->conf.output_sync[i]) {
                    synced_output_states sst;
                    switch(act) {
                        case outact_nop:
                            break;
                        case outact_on:
                            runfile_set_output(channel_id, true);
                            break;
                        case outact_off:
                            runfile_set_output(channel_id, false);
                            break;
                        case outact_switch:
                            sst = runfile_get_output(channel_id);
                            if(sst != synced_output_unset)
                                runfile_set_output(channel_id,
                                                   sst == synced_output_off);
                            break;
                    }
                }
                if(p->dev_hdl) {
                    // okay, performing now
                    do_perform_action(p, i, act);
                    return true;
                } else {
                    log_message("Can't control channel ``%s'': "
                                "the device %x is unavailable",
                                channel_id, p->conf.id);
                    return false;
                }
            }
    log_message("Output channel name ``%s'' not found, check config",
                channel_id);
    return false;
}

static bool dev_get_state(device_item *p)
{
    unc0xx_global_state_str state;
    int r = unc0xx_get_fullstate(p->dev_hdl, &state);
    if(r == -1) {
        device_lost(p);
        return false;
    }
    p->last_state = state.state_bits;
    return true;
}

static int output_state(int n, int state_bits, int version)
{
    if(n < 4)
        return (state_bits & (1<<n)) ? linestate_on : linestate_off;
    if(version >= 4000000) {
        if(n < 8)
            return (state_bits & (1<<(n+8))) ? linestate_on : linestate_off;
        else
            return (state_bits & (1<<n)) ? linestate_on : linestate_off;
    } else {
        // legacy
        return (state_bits & (1<<(n+4))) ? linestate_on : linestate_off;
    }
}

int devices_line_status(const char *line_id)
{
    int i;
    for(device_item *p = devices; p; p = p->next) {
        for(i=0; i<12; i++)
            if(p->conf.output[i] == line_id) {
                if(!p->dev_hdl)
                    return linestate_dev_lost;
                if(p->last_state == -1)
                    if(!dev_get_state(p))
                        return linestate_dev_lost;
                return output_state(i, p->last_state, p->version);
            }
        for(i=0; i<8; i++)
            if(p->conf.input_name[i] == line_id) {
                if(!p->dev_hdl)
                    return linestate_dev_lost;
                if(p->last_state == -1)
                    if(!dev_get_state(p))
                        return linestate_dev_lost;
                return (p->last_state & (1 << (i+8))) ?
                    linestate_on : linestate_off;
            }
    }
    log_message("Device line name ``%s'' not found, check config", line_id);
    return linestate_no_line;
}





DevicePoller::DevicePoller(SUEEventSelector *a_sel, int im, int rm, int osm)
    : selector(a_sel), interval(im),
    rescan_multiplier(rm), to_rescan(rm),
    outsync_multiplier(osm), to_outsync(osm)
{
    selector = a_sel;
    SetFromNow(0);
    selector->RegisterTimeoutHandler(this);
}

DevicePoller::~DevicePoller()
{
}

static void do_outsync_for(device_item *p)
{
    for(int i=0; i<12; i++) {
        if(p->conf.output[i].IsValid() && p->conf.output_sync[i]) {
            int lst = output_state(i, p->last_state, p->version);
                /* either linestate_on or linestate_off */
            int expected =
                runfile_get_output(p->conf.output[i].c_str());
            if(lst == linestate_on && expected == synced_output_off) {
                log_message("Warning: [%x] sync: line %d (%s) => off",
                            p->last_state,
                            i, p->conf.output[i].c_str());
                do_perform_action(p, i, outact_off);
            } else
            if(lst == linestate_off && expected == synced_output_on) {
                log_message("Warning: [%x] sync: line %d (%s) => on",
                            p->last_state,
                            i, p->conf.output[i].c_str());
                do_perform_action(p, i, outact_on);
            }
        }
    }
}

static void do_poll_inputs_for(device_item *p, int counters[8])
{
    for(int in = p->min_cnt; in <= p->max_cnt; in++) {
        if(counters[in] < 0 || counters[in] > 100) {
            // this is a workaround
            // it's unclear why the counter becomes garbaged sometimes 
            log_message("XXX: ignoring garbaged counter "
                        "in=%d l_i_s[]=%d(%x) cnt[]=%d(%x)",
                        in, p->last_in_state[in], p->last_in_state[in], 
                            counters[in], counters[in]);
            continue;
        }
        for(int c = p->last_in_state[in]+1; c<=counters[in]; c++) {
            ScriptVariable *ac = (c&1) ?
                &p->conf.input_on[in] : &p->conf.input_off[in];
            log_message("XXX: in=%d l_i_s[]=%d cnt[]=%d c=%d",
                        in, p->last_in_state[in], counters[in], c);
            if(ac->IsValid())
                state_process_event(ac->c_str());
        }
        p->last_in_state[in] = counters[in] & 1;
    }
}

static void do_poll_onewire_for(device_item *p)
{
    unc0xx_extra_state_str_common extra;
    do {
        int r = unc0xx_get_extrastate(p->dev_hdl, &extra);
        if(r == -1) {
            device_lost(p);
            break;
        }
        if(extra.page_version == 0) {
            unc0xx_extra_state_str_legacy *ex =
                (unc0xx_extra_state_str_legacy*)&extra; 
            for(int i=0; i<4; i++) {
                if(ex->ow_ids[i][0]) {
                    handle_onewire_info(p->conf.id, 
                        ex->ow_ids[i], ex->ow_temperature[i]);
                }
            }
        } else
        if(extra.page_version == 1) {
            unc0xx_extra_state_str_onewire *ex =
                (unc0xx_extra_state_str_onewire*)&extra; 
            for(int i=0; i<6; i++) {
                if(ex->ow_ids[i][0]) {
                    handle_onewire_info(p->conf.id, 
                        ex->ow_ids[i], ex->ow_temperature[i]);
                }
            }
        } else {
#if 0   // no, we'd better simply ignore this
        // okay, we don't use the 'delayed actions' page of the extra state
            log_message("Extra state vers. unknown (%d) for [%04x]",
                        extra.page_version, p->conf.id);
#endif
        }
    } while((extra.page_seq & 0x80));
}

void DevicePoller::TimeoutHandle()
{
    if(--to_rescan < 0)
        to_rescan = rescan_multiplier - 1;
    if(--to_outsync < 0)
        to_outsync = outsync_multiplier - 1;
    for(device_item *p = devices; p; p = p->next) {
        p->last_state = -1;
        if(!p->dev_hdl) {
            if(to_rescan)
                continue;
            connect_to_device(p, false);
            if(!p->dev_hdl)
                continue;
        }
        if(!to_outsync && p->output_sync_needed && dev_get_state(p)) {
            // dev_get_state updated the p->last_state field
            do_outsync_for(p);
        }
        if(p->poll_inputs) {
            int counters[8];
            int r = unc0xx_zero_query_counters(p->dev_hdl, counters, 8);
            if(r == -1) {
                device_lost(p);
                continue;
            }
            do_poll_inputs_for(p, counters);
        }
        if(p->conf.poll_onewire && p->version >= 2120203) {
            do_poll_onewire_for(p);
        }
    }
    SetFromNow(interval / 1000, (interval % 1000) * 1000);
        /* interval is in millisecs, SetFromNow takes secs and microsecs */
    selector->RegisterTimeoutHandler(this);
}

static void
devices_report_in(device_item *p, int in_n, bool bit, ScriptVariable &res)
{
    if(p->conf.input_name[in_n].IsValid()) {
        res += " ";
        res += p->conf.input_name[in_n];
        res += ":";
        res += bit ? "yes" : "no";
    }
}

static void
devices_report_out(device_item *p, int out_n, bool bit, ScriptVariable &res)
{
    if(p->conf.output[out_n].IsValid()) {
        res += " ";
        res += p->conf.output[out_n];
        res += ":";
        res += bit ? "on" : "off";
    }
}

/* get line number by the bit position within the 'status' word */
static int mod_out_number(int bit_n, int version)
{
    if(version >= 4000000) {
        return bit_n < 12 ? bit_n : bit_n - 8;
    } else {
        // legacy
        return bit_n - 4;
    }
}

ScriptVariable devices_report()
{
    ScriptVariable res;
    for(device_item *p = devices; p; p = p->next) {
        res += ScriptVariable(16, "[%04x:%s", p->conf.id,
                              p->dev_hdl ? "+" : " ?!");    
        if(!p->dev_hdl) {
            res += "]";
            continue;
        }
        unc0xx_global_state_str state;
        int r = unc0xx_get_fullstate(p->dev_hdl, &state);
        if(r == -1) {
            device_lost(p);
            continue;
        }
        int i;
        for(i = 0; i < 4; i++) { // main group, outputs only
            bool bit = (state.state_bits >> i) & 1;
            devices_report_out(p, i, bit, res);
        }
        for(i = 8; i < 16; i++) { // modules
            bool bit = (state.state_bits >> i) & 1;
            devices_report_out(p, mod_out_number(i, p->version), bit, res);
            devices_report_in(p, i-8, bit, res);
        }
        res += "] ";
    }
    return res;
}
