//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include "flog.hpp"

#define LOG_MSG_LIMIT 500

#ifndef WRITE_FLOG

void create_log(int version) {}
void log_message(const char *fmt, ...) {}
int log_fileno() { return -1; }
void log_watchdog_reset() {}

#else

#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

static FILE *log = 0;

static long msgcounter = 0;

void create_log(const char *fname)
{
    int fd = open(fname, O_WRONLY|O_CREAT|O_APPEND, 0666);
    log = fdopen(fd, "w"); 
}

void log_message(const char *fmt, ...)
{
    if(msgcounter == -1)
        return;
    char buf[64];
    time_t t = time(0);
    va_list vl;
    va_start(vl, fmt);
    strftime(buf, sizeof(buf), "%b %d %H:%M:%S ", localtime(&t));
    fprintf(log ? log : stderr, "%s [%ld] ", buf, msgcounter);
    if(++msgcounter > LOG_MSG_LIMIT) {
        fprintf(log ? log : stderr, "TOO MANY MESSAGES\n");
        msgcounter = -1;
    } else {
        vfprintf(log ? log : stderr, fmt, vl);
        fprintf(log ? log : stderr, "\n");
    }
    fflush(log ? log : stderr);
    va_end(vl);
}

void log_flush()
{
    fflush(log ? log : stderr);
}

int log_fileno()
{
    return log ? fileno(log) : -1;
}

void log_watchdog_reset()
{
    if(msgcounter == -1) {
        char buf[64];
        time_t t = time(0);
        strftime(buf, sizeof(buf), "%b %d %H:%M:%S ", localtime(&t));
        fprintf(log ? log : stderr, "%s", buf);
        fprintf(log ? log : stderr, "...RESUMING\n");
    }
    msgcounter = 0;
}


#endif
