//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <sue/sue_sel.hpp>

#include "config.hpp" // for sound_modes
#include "devices.hpp"
#include "lamp.hpp"

#define LNAME_LAMP  "^lamp"
#define LNAME_SOUND "^sound"

class LampBlinker : public SUETimeoutHandler {
    SUEEventSelector *selector;
    enum modes { off = -1, on = -2, blink_off = 0, blink_on = 1 } mode;
    int smode;
    int on_time, off_time;
public:
    LampBlinker(SUEEventSelector *a_sel)
        : selector(a_sel), mode(off), smode(snd_mode_off)
    {}
    ~LampBlinker() {}
    virtual void TimeoutHandle() {
        switch(mode) {
            case on:
                devices_output_action(LNAME_LAMP, outact_on);
                selector->RemoveTimeoutHandler(this);
                break;
            case off:
                devices_output_action(LNAME_LAMP, outact_off);
                selector->RemoveTimeoutHandler(this);
                break;
            case blink_off:
                devices_output_action(LNAME_LAMP, outact_on);
                mode = blink_on;
                HandleSoundBlink();
                SetFromNow(on_time / 1000, (on_time % 1000) * 1000);
                selector->RegisterTimeoutHandler(this);
                break;
            case blink_on:
                devices_output_action(LNAME_LAMP, outact_off);
                mode = blink_off;
                HandleSoundBlink();
                SetFromNow(off_time / 1000, (off_time % 1000) * 1000);
                selector->RegisterTimeoutHandler(this);
                break;
        }
    }
    void SetMode(int light, int pause, int sm) {
        if(sm != snd_mode_unchanged) {
            smode = sm;
            switch(sm) {
                case snd_mode_on:
                    devices_output_action(LNAME_SOUND, outact_on);
                    break;
                case snd_mode_off:
                    devices_output_action(LNAME_SOUND, outact_off);
                    break;
                default:
                    ;
            }
        }
        if(light != -1 && pause != -1 &&
            (light != on_time || pause != off_time))
        {
            on_time = light;
            off_time = pause;
            if(light == 0) mode = off;
            else
            if(pause == 0) mode = on;
            else
            if(mode == off) mode = blink_off;
            else
            if(mode == on) mode = blink_on;
            selector->RemoveTimeoutHandler(this);
            TimeoutHandle();
        }
    }
private:
    void HandleSoundBlink() {
        if(smode == snd_mode_off || smode == snd_mode_on) return;
        bool seton = 
            (mode == blink_on  && smode == snd_mode_withlamp) ||
            (mode == blink_off && smode == snd_mode_against);
        devices_output_action(LNAME_SOUND, seton ? outact_on : outact_off);
    }
};

static LampBlinker *blinker = 0;

void lamp_init(class SUEEventSelector *selector)
{
    if(blinker) delete blinker;
    blinker = new LampBlinker(selector);
}

    // numbers are in milliseconds, that is, 1000 is 1 second
    // special combinations: (1, 0) lamp is on, (0, 0) lamp is off
void lamp_set(int light, int pause, int sound)
{
    if(!blinker) return;
    blinker->SetMode(light, pause, sound);
}
