//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <scriptpp/scrvar.hpp>
#include <sue/sue_sel.hpp>
#include <sue/sue_wait.hpp>

#include "flog.hpp"
#include "config.hpp"
#include "runfile.hpp"
#include "state.hpp"
#include "action.hpp"
#include "devices.hpp"
#include "onewire.hpp"
#include "console.hpp"
#include "webserv.hpp"
#include "smscmd.hpp"
#include "watchdog.hpp"

#ifndef FFAZENDA_VERSION
#define FFAZENDA_VERSION "unknown_version"
#endif

#ifndef THE_USER
#define THE_USER "unknown_person"
#endif

static void help()
{
    static const char *htext[] = {
        "ffazenda vers. " FFAZENDA_VERSION
            " (c) Unicontrollers Ltd., 2013-2014",
        "usage: ffazenda <options>",
        "  options:  -c <file>       set the configuration file",
        "            -l <file>       set the log file",
        "            -f              (foreground) don't daemonize",
        0        
    };
    int i;
    for(i=0; htext[i]; i++)
        printf("%s\n", htext[i]);
}


SUEChildWaitAgent *the_child_agent = 0;

static ScriptVariable scv_get_cwd()
{
    for(int len = 32; ; len += 2) {
        char *buf = new char[len];
        if(getcwd(buf, len)) {
            ScriptVariable res(buf);
            delete[] buf;
            return res;
        }
        delete[] buf;
    }
}

static void mass_close()
{
    int i;
    int d = getdtablesize();
    for(i = 3; i < d; i++)
        close(i);
}

int main(int argc, char **argv)
{
    ScriptVariable config_fname("ffazenda.ini");
    bool daemonize = true;
    bool log_opened = false;
    const char * const * p = argv+1;
    while(*p) {
        if(ScriptVariable(*p) == "-c") {
            if(*(p+1))
                config_fname = *(p+1);
            else {
                help();
                log_message("``-c'' requires parameter");
                return 1;
            }
            p++;
        } else
        if(ScriptVariable(*p) == "-l") {
            if(!log_opened) {
                if(*(p+1)) {
                    create_log(*(p+1));
                    log_opened = true;
                } else {
                    help();
                    log_message("``-l'' requires parameter");
                    return 1;
                }
            } else {
                help();
                log_message("duplicate ``-l''");
                return 1;
            }
            p++;
        } else
        if(ScriptVariable(*p) == "-f") {
            daemonize = false;
        } else {
            help();
            log_message("unknown key ``%s''", *p);
            return 1;
        }
        p++;
    }

    if(!init_config(config_fname.c_str())) {
        return 1;
    }

    if(!log_opened) {
        ScriptVariable s = config_get_log_name();
        create_log(s.c_str());
    }

    log_message("START=> FFazenda vers. " FFAZENDA_VERSION " compiled at "
                __DATE__ " " __TIME__ " by " THE_USER);
    config_report();

    runfile_init(config_get_runfile_name().c_str());

    ScriptVariable my_command = argv[0];
    if(my_command[0] != '/' && my_command.Strchr('/').IsValid()) {
        // it's a relative path, we need to memorize the binary as
        // our working dir will be '/' once we daemonize
        ScriptVariable Cwd = scv_get_cwd();
        my_command = Cwd + '/' + my_command;
    }

    if(daemonize) {
        if(-1 == chdir("/")) {
            perror("warning: /");
        }
        int fd = open("/dev/null", O_RDWR);
        dup2(fd, 0); 
        dup2(fd, 1); 
        dup2(fd, 2); 
        if(fork() > 0) exit(0);
        setsid();
    }

    init_onewire_info(); // must be after the config but before the devices

    SUEEventSelector  selector;
    SUEChildWaitAgent child_agent;
    child_agent.Register(&selector);
    the_child_agent = &child_agent;

    devices_init(&selector);  // must be before state_init
    state_init(&selector);
    log_message("starting in the mode %s", get_current_state());

    init_onewire_state(); // must be after the state and the devices

    DevicePoller poller(&selector, 500 , 5, 13);
         /* 500 = 2 times a second, 5 = rescan after 2.5 sec,
                        13 = sync outputs after 7.5 sec */
    console_init(&selector);
    webserv_init(&selector);
    SmscmdPoller smspoller(&selector, 1000 /* once a second */);

    int watchdog_interval = config_get_watchdog_interval();
    if(watchdog_interval) {
        new Watchdog(&selector, watchdog_interval * 1000);
    }

    action_memorize_selector(&selector);
    selector.Go();
    if(action_should_restart()) {
        argv[0] = const_cast<char*>(my_command.c_str());
        log_flush();
        mass_close();
        execvp(argv[0], argv);
        log_message("Restarting (with argv[0]==%s) failed", argv[0]);
        exit(1);
    }
    return 0;
}
