//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <unistd.h> // for getcwd
#include <inifile/inifile.hpp>

#include "flog.hpp"
#include "config.hpp"

#include "runfile.hpp"

static IniFileParser *run = 0;
static ScriptVariableInv run_fname;

void runfile_init(const char *fname)
{
    if(run)
        delete run;
    run = new IniFileParser();
    if(*fname == '/') {
        run_fname = fname;
    } else {
        char buf[1024];
        buf[0] = 0;
        if(getcwd(buf, sizeof(buf))) {
            run_fname = buf;
            run_fname += '/';
            run_fname += fname;
        } else {
            run_fname = fname;
        }
    }
    if(!run->Load(run_fname.c_str())) {
        log_message("%s:%d: warning: reading state file problem: %s",
                    fname,
                    run->GetLastErrorLine(),
                    run->GetLastErrorDescription());
    }
}

ScriptVariable runfile_get_state()
{
    ScriptVariableInv invalid;
    if(!run)
        return invalid;
    const char *res = run->GetTextParameter("State", 0, "alert", 0);
    if(!res)
        return invalid;
    return res;
}

void runfile_set_state(const char *state)
{
    if(!run || run_fname.IsInvalid())
        return;
    run->SetTextParameter("State", 0, "alert", state);
    run->Save(run_fname.c_str());
}

static ScriptVariable tempwm_param_name(int index)
{
    return ScriptVariable(16, "wmark_%d", index);
}

ScriptVariable runfile_get_tempwm(const char *name, int index)
{
    if(!run)
        return "undef";
    ScriptVariable pn = tempwm_param_name(index);
    return run->GetTextParameter("TempSensor", name, pn.c_str(), "undef");
}

void runfile_set_tempwm(const char *name, int index, const char *v)
{
    if(!run || run_fname.IsInvalid())
        return;
    ScriptVariable pn = tempwm_param_name(index);
    run->SetTextParameter("TempSensor", name, pn.c_str(), v);
    run->Save(run_fname.c_str());
}

ScriptVariable runfile_get_variable(const char *name)
{
    ScriptVariableInv inv;
    if(!config_runvar_exists(name))
        return inv;
    if(run) {
        const char *val = run->GetTextParameter("Variables", 0, name, 0);
        if(val)
            return val;
    }
    // okay, it's not there, what about the default?
    return config_runvar_default(name);
}

bool runfile_set_variable(const char *name, const char *value)
{
    if(!run || run_fname.IsInvalid())
        return false;
    if(!config_runvar_exists(name))
        return false;
    run->SetTextParameter("Variables", 0, name, value);
    run->Save(run_fname.c_str());
    return true;
}

synced_output_states runfile_get_output(const char *name)
{
    if(!run)
        return synced_output_unset;
    const char *res =
        run->GetTextParameter("Outputs", 0, name, 0);
    if(!res)
        return synced_output_unset;
    return ScriptVariable(res).Trim().Tolower() == "on" ?
        synced_output_on : synced_output_off;
}

void runfile_set_output(const char *name, bool st)
{
    run->SetTextParameter("Outputs", 0, name, st ? "on" : "off");
    run->Save(run_fname.c_str());
}
