//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <stdio.h> // for rename(2)
#include <string.h> // for strerror
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <sue/sue_sel.hpp>
#include <scriptpp/cmd.hpp>

#include "config.hpp"
#include "flog.hpp"
#include "smsg.hpp"
#include "commands.hpp"

#include "smscmd.hpp"

#define OUTG_PREFIX "FFAZENDA_OUTG"

/* WARNING

This module is implemented 'ad hoc' for the very configuraiton of smsd
which is actually on the machine we use.

It is almost certain that it will NOT work in your configuration!

The function to change is 'do_polling()'; it makes too many assumptions!

 */

static bool send_sms(const CommonMessage &message)
{
    const char *outg = config_get_sms_outgoing_dir();
    if(!outg) return false;
    const char *compos = config_get_sms_composing_dir();
    if(!compos) return false;
    static int id = 0;
    int t = time(0);
    id += (t%1000);
    ScriptVariable fname(0, OUTG_PREFIX ".%d.%d.%d", getpid(), t, id);

    ScriptVariable to_hdr = message.FindHeader("To");
    if(to_hdr.IsInvalid()) {
        log_message("Attempt to sent sms to nowhere");
        return false;
    }
    ScriptVariable body_str = message.GetBodyStr();
    log_message("Sending sms %s to [%s]: [%s]",
                fname.c_str(), to_hdr.c_str(), body_str.c_str());

    ScriptVariable srcfile = ScriptVariable(compos) + "/" + fname;
    ScriptVariable dstfile = ScriptVariable(outg) + "/" + fname;
    FILE *f = fopen(srcfile.c_str(), "w");
    if(!f) {
        log_message("Couldn't open ``%s'' for writing: %s",
                    srcfile.c_str(), strerror(errno));
        return false;
    }
    message.OutputMessage(f);
    fclose(f);
    int r = rename(srcfile.c_str(), dstfile.c_str());
    if(r == -1) {
        log_message("Couldn't rename ``%s'' to ``%s'': %s", 
                    srcfile.c_str(), dstfile.c_str(),
                    strerror(errno));
        return false;
    }
    return true;
}

bool send_sms_to(const ScriptVariable &number, const ScriptVariable &message)
{
    CommonMessage msg;
    msg.AddHeader("To", number);
    msg.AddStringToBody(message.c_str());
    return send_sms(msg);
}

static bool incoming_sms_name_check(const ScriptVariable &name)
{
    // we only take files matching 'GSM1.??????'
    //   and ignore everything else
    //   this is no good, actually....
    return name.HasPrefix("GSM1.") && name.Length() == 11;
}

static void process_sms(const ScriptVariable &filename)
{
    CommonMessage msg;
    FILE *fl = fopen(filename.c_str(), "r");
    if(!fl) {
        log_message("Couldn't open ``%s'' for reading, aborting: %s",
                    filename.c_str(), strerror(errno));
        return;
    }
    int c;
    while((c = fgetc(fl)) != EOF) {
        msg.FeedByte(c);
    }
    fclose(fl);

    ScriptVariable from = msg.FindHeader("From");
    if(from.IsInvalid()) {
        log_message("SMS %s contained no From: field, "
                    "will process as from hostile world",
                    filename.c_str());
        from = "";
    }

    ScriptVariable conmsg("");
    ScriptVariable charset = msg.FindHeader("Alphabet");
    charset.Trim();
    if(charset.IsValid() && charset == "UCS2") {
        const unsigned char *b = msg.GetBody();
        for(int i=0; i<msg.GetBodyLength()-1; i+=2) {
            if(b[i] == 0) {
                if(isprint(b[i+1])) {
                    conmsg += (char)b[i+1];
                } else
                if(isspace(b[i+1])) {
                    conmsg += ' ';
                } else
                    conmsg += '?';
            } else
                conmsg += '?';
        }
    } else {
        const unsigned char *b = msg.GetBody();
        for(int i=0; i<msg.GetBodyLength(); i++) {
            conmsg += isprint(b[i]) ? (char)b[i] :
                          (isspace(b[i]) ? ' ' : '?');
        }
    }
    log_message("Message from [%s] contained: ``%s''",
                from.c_str(), conmsg.c_str());
    ScriptVariable response;
    response_status rs = process_command(from, conmsg, response);
    if(rs == response_important) {
        if(!send_sms_to(from, response)) {
            log_message("Couldn't respond, sorry...");
        }
    }
}

static bool failed_sms_name_check(const ScriptVariable &name)
{
    if(!name.HasPrefix(OUTG_PREFIX "."))
        return false;
    if(!name.HasSuffix(".LOCK"))
        return false;
    return true;
}

static void resend_sms(const ScriptVariable &filename)
{
    CommonMessage msg;
    FILE *fl = fopen(filename.c_str(), "r");
    if(!fl) {
        log_message("Couldn't open ``%s'' for reading, aborting: %s",
                    filename.c_str(), strerror(errno));
        return;
    }
    int c;
    while((c = fgetc(fl)) != EOF) {
        msg.FeedByte(c);
    }
    fclose(fl);

    ScriptVariable to_hdr = msg.FindHeader("To");
    ScriptVariable fail_date = msg.FindHeader("Failed");
    ScriptVariable fail_reason = msg.FindHeader("Fail_reason");
    ScriptVariable attempt_hdr = msg.FindHeader("Attempt");
    long attempt;

    if(to_hdr.IsInvalid()) {
        log_message("Found failed sms [%s] without ``To:'' header, skipping",
                    filename.c_str());
        return;
    }
    if(fail_date.IsInvalid()) fail_date = "unknown";
    if(fail_reason.IsInvalid()) fail_reason = "unknown";
    if(attempt_hdr.IsInvalid())
        attempt = 1;
    else
        if(!attempt_hdr.GetLong(attempt))
            attempt = 1;

    log_message("Found failed sms [%s], to %s, date %s, reason %s, att. %d",
                filename.c_str(), to_hdr.c_str(),
                fail_date.c_str(), fail_reason.c_str(), attempt);
    if(attempt > 5) { // XXX hardcoded value 5 to be changed...
        log_message("Tried too many times, skipping");
        return;
    }
    attempt++;
    msg.ClearHeaders();    
    msg.AddHeader("To", to_hdr);
    msg.AddHeader("Attempt", ScriptNumber(attempt));
    send_sms(msg);
}

static bool lock_file_exists(const ScriptVariable &basename)
{
    return FileStat((basename + ".LOCK").c_str()).Exists();
}

static void traverse_sms_dir(const char *incom,
                             const char *compos,
                             const char *done, 
                             bool (*check_fname)(const ScriptVariable &),
                             void (*proc_func)(const ScriptVariable&))
{
    ReadDir rd(incom);
    const char *dir;
    while((dir = rd.Next())) {
        ScriptVariable ds(dir);
        if(!check_fname(ds))
            continue;
        // okay, let's go...
        ScriptVariable srcfile = ScriptVariable(incom) + "/" + ds;
        if(lock_file_exists(ds)) {
            log_message("File %s locked, skipping", ds.c_str());
            continue;
        }

        ScriptVariable dstfile = ScriptVariable(compos) + "/" + ds;
        int r = rename(srcfile.c_str(), dstfile.c_str());
        if(r == -1) {
            log_message("Couldn't rename ``%s'' to ``%s'', ignoring: %s", 
                        srcfile.c_str(), dstfile.c_str(),
                        strerror(errno));
            continue;
        }
        proc_func(dstfile);
        if(!done) continue;
        ScriptVariable donefile = ScriptVariable(done) + "/" + ds;
        r = rename(dstfile.c_str(), donefile.c_str());
        if(r == -1)
            log_message("Warning: couldn't rename ``%s'' to ``%s'', "
                        "ignoring: %s", 
                        dstfile.c_str(), donefile.c_str(),
                        strerror(errno));
    }

}

static void do_polling()
{
    const char *incom = config_get_sms_incoming_dir();
    const char *compos = config_get_sms_composing_dir();
    const char *done = config_get_sms_proceeded_dir();
    const char *failed = config_get_sms_failed_dir();
    if(incom && compos) {
        traverse_sms_dir(incom, compos, done,
                         incoming_sms_name_check, process_sms);
    }
    if(failed && compos) {
        traverse_sms_dir(failed, compos, done,
                         failed_sms_name_check, resend_sms);
    }
}

SmscmdPoller::SmscmdPoller(SUEEventSelector *a_sel, int interv_ms)
    : selector(a_sel), interval(interv_ms)
{
    selector = a_sel;
    SetFromNow(0);
    selector->RegisterTimeoutHandler(this);
}

SmscmdPoller::~SmscmdPoller()
{
    selector->RemoveTimeoutHandler(this);
}

void SmscmdPoller::TimeoutHandle()
{
    do_polling();
    SetFromNow(interval / 1000, (interval % 1000) * 1000);
    selector->RegisterTimeoutHandler(this);
}

