//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <scriptpp/scrvar.hpp>
#include <scriptpp/scrvect.hpp>

#include "smsg.hpp"

CommonMessage::CommonMessage()
    : headers(), values(), buf(0), buf_len(0), buf_used(0), state(Eol)
{}

CommonMessage::~CommonMessage()
{
    if(buf) delete[] buf;
}

void CommonMessage::Clear()
{
    ClearHeaders();
    ClearBody();
    state = Eol;
}

void CommonMessage::ClearHeaders()
{
    headers.Clear();
    values.Clear();
}

void CommonMessage::ClearBody()
{
    if(buf) delete[] buf;
    buf = 0;
    buf_len = buf_used = 0;
}

void CommonMessage::FeedByte(unsigned char b)
{
    int lh = headers.Length();
    switch(state) {
        case Name:
            if(b == '\n') {
                values[lh-1] = ScriptVariableInv();
                state = Eol;
                break;
            } else
            if(b == ':') {
                state = Colon;
                break;
            }
            headers[lh-1] += (char)b;
            break;
        case Colon:
            state = Value;
            if(b == ' ')
                break;
            // otherwise it is an error but let's be liberal :-)
            // intentionally no break here!
        case Value:
            if(b != '\n') {
                values[lh-1] += (char)b;
                break;
            }
            state = Eol;
            break;
        case Eol:
            if(b == '\n') {
                state = Body;
            } else {
                state = Name;
                ScriptVariable s("");
                s += (char)b;
                headers.AddItem(s);
                values.AddItem(ScriptVariable(""));
            }
            break;
        case Body:
            AddByteToBody(b);
    }
}

void CommonMessage::
AddHeader(const ScriptVariable &head, const ScriptVariable &val)
{
    while(headers.Length() < values.Length()) headers.AddItem("");
    while(headers.Length() > values.Length()) values.AddItem("");
    headers.AddItem(head);
    values.AddItem(val);
}

void CommonMessage::AddByteToBody(unsigned char b)
{
    ResizeBody(buf_used + 1);
    buf[buf_used++] = b;
}

void CommonMessage::AddStringToBody(const char *str)
{
    while(*str) AddByteToBody(*str++);
}

int CommonMessage::GetHeadersCount() const
{
    return headers.Length();
}

bool CommonMessage::
GetHeader(int idx, ScriptVariable &name, ScriptVariable &val) const
{
    if(idx >= headers.Length()) return false;
    name = headers[idx];
    val = values[idx];
    return true;
}

ScriptVariable CommonMessage::FindHeader(const ScriptVariable& name) const
{
    for(int i=0; i<headers.Length(); i++)
        if(headers[i] == name) return values[i];
    return ScriptVariableInv();
}

void CommonMessage::
SetHeader(const ScriptVariable &name, const ScriptVariable &val)
{
    for(int i=0; i<headers.Length(); i++)
        if(headers[i] == name) {
            values[i] = val;
            return;
        }
    // not found, adding
    AddHeader(name, val);
}

ScriptVariable CommonMessage::GetBodyStr() const
{
    ScriptVariable res;
    for(int k = 0; k < buf_used; k++)
        res += (buf[k] != 0) ? buf[k] : '?';
    return res; 
}

void CommonMessage::OutputMessage(FILE *f) const
{
    for(int i=0; i<headers.Length(); i++)
        fprintf(f, "%s: %s\n", headers[i].c_str(), values[i].c_str());
    if(buf_used > 0) {
        fprintf(f, "\n");
        for(int k = 0; k < buf_used; k++)
            fputc(buf[k], f);
    }
}

void CommonMessage::ResizeBody(int atleast)
{
    if(atleast<buf_len) return;
    if(buf_len == 0) {
        int nl = 1;
        while(nl < atleast) nl *= 2;
        buf = new unsigned char[nl];
        buf_len = nl;
    } else {
        int nl = buf_len;
        while(nl < atleast) nl *= 2;
        unsigned char *nb = new unsigned char[nl];
        for(int i = 0; i<buf_used; i++) nb[i] = buf[i];
        delete[] buf;
        buf = nb;
        buf_len = nl;
    }
}
