//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <scriptpp/scrvar.hpp>
#include <scriptpp/scrvect.hpp>
#include <scriptpp/scrsubs.hpp>
#include <scriptpp/scrmap.hpp>
#include <sue/sue_sel.hpp>

#include "flog.hpp"
#include "lamp.hpp"
#include "config.hpp"
#include "runfile.hpp"
#include "action.hpp"
#include "evexec.hpp"
#include "notify.hpp"
#include "textsubs.hpp"
#include "state.hpp"

static SUEEventSelector *selector;

class StatesSet : public ScriptMap {
public:
    StatesSet() : ScriptMap() {}
    StatesSet(const ScriptVariable &s);
    ScriptVariable Join() const;
};

StatesSet::StatesSet(const ScriptVariable &s)
{
    ScriptWordVector v(s);
    for(int i = 0; i < v.Length(); i++) {
        ScriptTokenVector t(v[i], ".");
        if(t.Length() > 1)
            (*this)[t[0]] = t[1];
        else
            (*this)[""] = t[0];
    } 
}

ScriptVariable StatesSet::Join() const
{
    ScriptVariable key, val, res;
    Iterator iter(*this);
    while(iter.GetNext(key, val)) {
        if(res != "") res += " ";
        if(key != "") {
            res += key;
            res += ".";
        }
        res += val;
    }
    return res;
}

static StatesSet current_states;

static void sync_run()
{
    runfile_set_state(get_current_state());
}

static void go_to_mode(const ScriptVariable &newmode, bool suppress_notif);

void state_init(SUEEventSelector *sel)
{
    selector = sel;
    lamp_init(selector);
    current_states[""] = "";
    ScriptVariable st = runfile_get_state();
    if(st.IsInvalid()) {
        st = config_get_defaultstate();
    }
    ScriptWordVector st_vec(st);
    for(int ii = 0; ii < st_vec.Length(); ii++)
        go_to_mode(st_vec[ii], true); // no state notifications on start!
}

const char *get_current_state()
{
    static ScriptVariable r;
    r = current_states.Join();
    return r.c_str();
}

const char *get_current_state_of(const char *automata_id)
{
    return current_states[automata_id].c_str();
}

class TimeoutTransit : public SUETimeoutHandler {
    static TimeoutTransit *first;
    TimeoutTransit *next;

    SUEEventSelector *selector;
    int seconds;
    ScriptVariable newmode, event;
    ScriptVariable fsm;
public:
    TimeoutTransit(SUEEventSelector *a_sel, int sec,
                   const ScriptVariable &nm, const ScriptVariable &evt)
        : selector(a_sel), seconds(sec), newmode(nm), event(evt)
    {
        ScriptVariable::Substring ss = newmode.Strstr(".");
        if(ss.IsInvalid())
            fsm = "";
        else
            fsm = ss.Before().Get();
        CancelFsm(fsm);
        next = first;
        first = this;
        SetFromNow(seconds);
        selector->RegisterTimeoutHandler(this);
    }
    ~TimeoutTransit() {
        for(TimeoutTransit **t = &first; *t; t = &(*t)->next)
            if(*t == this) {
                *t = next;
                next = 0;
                break;
            }
    }
    virtual void TimeoutHandle() {
        ScriptVariable nm = newmode;
        ScriptVariable ev = event;
        delete this;
        // This way we guarantee that the object is already deleted so
        // it will not be deleted second time because of cancelling,
        // which may follow as a result of changing the state and/or
        // processing the event.
        // However, THE OBJECT IS NO LONGER HERE! so don't access the
        // fields

        if(nm.IsValid() && nm != "")
            go_to_mode(nm, false);
        if(ev.IsValid() && ev != "")
            state_process_event(ev.c_str());
    }
    void Cancel() {
        log_message("Cancelling timeout-based tranfer to %s", newmode.c_str());
        selector->RemoveTimeoutHandler(this);
        delete this;
    }
    static void CancelFsm(const ScriptVariable &fsm) {
        for(TimeoutTransit *t = first; t; t = t->next) {
            if(t->fsm == fsm) {
                t->Cancel();
                break;
            }
        }
    }
};

TimeoutTransit* TimeoutTransit::first = 0;

// only changes a single mode, the mode name should be qualified,
// or else the main FSM assumed
static void go_to_mode(const ScriptVariable &nm, bool suppress_notif)
{
    int lamp_on, lamp_off;
    sound_modes snd;
    ScriptVariable notific;
    int trans_tm;
    ScriptVariable newmode = nm, trans_mode, trans_event, unready_mode;
    if(!config_get_state_properties(newmode.c_str(),
        lamp_on, lamp_off, snd, notific,
        trans_tm, trans_mode, trans_event, unready_mode))
    {
        log_message("conf. error for %s; not changing to it",
                    newmode.c_str());
        return;
    }

    if(!ready_for_security_now() &&
        unready_mode.IsValid() && unready_mode != ""
    ) {
        log_message("not ready for security; going to %s instead of %s",
                    unready_mode.c_str(), newmode.c_str());
        ScriptVariable ignored;
        if(!config_get_state_properties(unready_mode.c_str(),
            lamp_on, lamp_off, snd, notific,
            trans_tm, trans_mode, trans_event, ignored))
        {
            log_message("conf. error for %s; not changing to it",
                        unready_mode.c_str());
            return;
        }
        newmode = unready_mode;
    }

    log_message("YYY newstate %s lamp/sound %d:%d/%d notif [%s] trans [%d:%s]",
                newmode.c_str(), lamp_on, lamp_off, snd, notific.c_str(),
                trans_tm, trans_mode.c_str());

    ScriptVariable::Substring ss = newmode.Strstr(".");
    ScriptVariable fsm =
        ss.IsInvalid() ? ScriptVariable("") : ss.Before().Get();
    TimeoutTransit::CancelFsm(fsm);

    ScriptVector dict; dict.AddItem("newstate"); dict.AddItem(newmode);
    ScriptVector argv; // empty

    if(notific.IsValid() && !suppress_notif) {
        ScriptVariable nn = FFSubstitution(notific, argv, dict);
        perform_notification(nn.c_str());
    }
    lamp_set(lamp_on, lamp_off, snd);

    ScriptTokenVector nmvec(newmode, ".", " \t");
    if(nmvec.Length() < 2)
        nmvec.Insert(0, "");

    current_states[nmvec[0]] = nmvec[1];

    log_message("STATE is now ``%s''", current_states.Join().c_str());
    sync_run();

    if(trans_tm != -1) {
        ScriptVariable trev = FFSubstitution(trans_event, argv, dict);
        log_message("YYY scheduling transfer to %s in %d",
                    trans_mode.c_str(), trans_tm);
        new TimeoutTransit(selector, trans_tm, trans_mode, trans_event);
    }
}

static void process_event_action(const char *ac, const ScriptVector& argv)
{
    ScriptVector ac_vec(ac, "\n", " \r\t");
    for(int i = 0; i < ac_vec.Length(); i++) {
        if(ac_vec[i] == "")
            continue;
        ScriptVariable action = FFSubstitution(ac_vec[i], argv);
        perform_action(action.c_str());
    }
}

#if 0
static void make_notification_sub(ScriptSubstitution &s,
                      const ScriptVector &argv, const char *newstate)
{
    s.AddVar(new ScriptSubstitutionConst("state", get_current_state()));
    s.AddVar(new ScriptSubstitutionConst("newstate", newstate));
    s.AddVar(new ScriptSubstitutionArgv(&argv, false));
}
#endif

static void do_transit_on_event(const char *transit_table)
{
    ScriptTokenVector trt(transit_table, "\n", " \t\r");
    StatesSet new_states;
    for(int i = 0; i < trt.Length(); i++) {
        // fsm.mode1 -> fsm.mode2
        if(trt[i] == "")
            continue;
        ScriptVariable::Substring arrow = trt[i].Strstr("->");
        if(arrow.IsInvalid()) {
            log_message("conf. error: no ``->'' in %s", trt[i].c_str());
            continue;
        }
        ScriptVariable from_m = arrow.Before().Get().Trim();
        ScriptTokenVector from_m_vec(from_m, ".", " \t");
        if(from_m_vec.Length() < 2)
            from_m_vec.Insert(0, "");
        ScriptVariable to_m = arrow.After().Get().Trim();
        ScriptTokenVector to_m_vec(to_m, ".", " \t");
        if(to_m_vec.Length() < 2)
            to_m_vec.Insert(0, "");
        if(from_m_vec[0] != to_m_vec[0]) {
            log_message("conf. error: %s and %s aren't from one fsm",
                        from_m.c_str(), to_m.c_str());
            continue;
        }
        if(current_states[from_m_vec[0]] == from_m_vec[1])
            new_states[from_m_vec[0]] = to_m_vec[1];
        else
        if(from_m_vec[1] == "*" && !new_states.Contains(from_m_vec[0]))
        {
            new_states[from_m_vec[0]] = to_m_vec[1];
        }
    }
    if(new_states.Count() > 0)
        log_message("new states: %s", new_states.Join().c_str());
    ScriptVariable fsm, nst;
    StatesSet::Iterator iter(new_states);
    while(iter.GetNext(fsm, nst)) {
        go_to_mode(fsm == "" ? nst : (fsm + "." + nst), false);
    }
}

void state_process_event(const char *event)
{
    log_message("Processing event ``%s''", event);

    ScriptVector argv(event);

    const char *fsm = config_get_event_fsm(argv[0].c_str());
    ScriptVariable selst = current_states[fsm];

    const char *act = config_get_event_action(argv[0].c_str(), selst.c_str());
    if(act && *act) {
        log_message("Doing (event) action ``%s''", act);
        process_event_action(act, argv);
    }

#if 1  // it's not obvious whether this is a good idea or not...
    const char *execln =
        config_get_event_exec(argv[0].c_str(), selst.c_str());
    if(execln && *execln)
        process_event_exec(execln, argv);
#endif

    const char *notification =
        config_get_event_notification(argv[0].c_str(), selst.c_str());
    if(notification) {
        // ScriptVector dict; dict.AddItem("newstate"); dict.AddItem(newstate);
        ScriptVariable nn = FFSubstitution(notification, argv /*,dict*/);
        perform_notification(nn.c_str());
    }

    const char *transit_table =
        config_get_event_transit(argv[0].c_str(), selst.c_str());
    if(transit_table && *transit_table) {
        do_transit_on_event(transit_table);
    }

    const char *subevents =
        config_get_event_subevents(argv[0].c_str(), selst.c_str());
    if(subevents && *subevents) {
        ScriptVariable sub = FFSubstitution(subevents, argv /*,dict*/);
        ScriptVector sub_vec(sub, "\n", " \t\r");
        for(int i = 0; i < sub_vec.Length(); i++) {
            if(sub_vec[i] != "")
                state_process_event(sub_vec[i].c_str());
        }
    }
}
