//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <time.h>

#include "config.hpp"
#include "runfile.hpp"
#include "state.hpp"
#include "devices.hpp"
#include "onewire.hpp"

#include "textsubs.hpp"


// note the class doesn't own these ScritpVectors (both of them!)
// make sure they continue to exist at least as long as the object
// of DictionaryVar exists.
class DictionaryVar : public ScriptSubstitutionArgv {
    const ScriptVector *the_dict;
public:
    DictionaryVar(const ScriptVector &argv, const ScriptVector &dict)
        : ScriptSubstitutionArgv(&argv, false), the_dict(&dict)
    {}
    ScriptVariable Sub(const char *name, int len) const
    {
        ScriptVariable v = ScriptSubstitutionArgv::Sub(name, len);
        if(v.IsValid()) return v;
        for(int i = 0; i < the_dict->Length() - 1; i += 2) {
            if((*the_dict)[i].Length() == len) {
                bool ok = true;
                for(int k = 0; k < len; k++)
                    if(name[k] != (*the_dict)[i][k]) {
                        ok = false;
                        break;
                    }
                if(ok)
                    return (*the_dict)[i+1];
            }
        }
        return ScriptVariableInv();
    }
};

class SubCurrentState : public ScriptSubstitutionSingleName {
public:
    SubCurrentState() : ScriptSubstitutionSingleName("state") {}
    virtual ScriptVariable Text() const
        { return get_current_state(); }
};

class SubSystemName : public ScriptSubstitutionSingleName {
public:
    SubSystemName() : ScriptSubstitutionSingleName("sysname") {}
    virtual ScriptVariable Text() const
        { return config_get_system_name(); }
};

class SubDevReport : public ScriptSubstitutionSingleName {
public:
    SubDevReport() : ScriptSubstitutionSingleName("devices_report") {}
    virtual ScriptVariable Text() const
        { return devices_report(); }
};

class SubOWReport : public ScriptSubstitutionSingleName {
public:
    SubOWReport() : ScriptSubstitutionSingleName("onewire_report") {}
    virtual ScriptVariable Text() const
        { return onewire_report(); }
};

class SubDateReport : public ScriptSubstitutionSingleName {
public:
    SubDateReport() : ScriptSubstitutionSingleName("date") {}
    virtual ScriptVariable Text() const {
        time_t t = time(0);
        const struct tm *ptm = localtime(&t);
        ScriptVariable res(16, "%d%02d%02d %02d:%02d",
            ptm->tm_year+1900, ptm->tm_mon+1, ptm->tm_mday,
            ptm->tm_hour, ptm->tm_min
        );
        return res;
    }
};

class SubFullReport : public ScriptSubstitutionSingleName {
public:
    SubFullReport() : ScriptSubstitutionSingleName("full_report") {}
    virtual ScriptVariable Text() const {
        return ScriptVariable("status: ") + get_current_state() 
            + devices_report() + onewire_report();
    }
};

class SubRunVariables : public ScriptSubstitution::Var {
public:
    SubRunVariables() {}
    ScriptVariable Sub(const char *name, int len) const {
        if(*name != '?')
            return ScriptVariableInv();
        ScriptVariable vn;
        for(int i = 1; i < len; i++)
            vn += name[i];
        return runfile_get_variable(vn.c_str());
    }
};

class SubDevLines : public ScriptSubstitutionPrefixRequest {
public:
    SubDevLines() : ScriptSubstitutionPrefixRequest("line") {}
    virtual ScriptVariable Handle(ScriptVector &params) const  {
        // line:the_line_name:on:off:n/a
        //  where the_line_name is either input or output,
        //  on:off:n/a -- messages to return
        int r = devices_line_status(params[0].c_str());
        switch(r) {
            case linestate_on:       return params[1];
            case linestate_off:      return params[2];
            default:                 return params[3];
        }
    }
};

class SubTemperature : public ScriptSubstitutionPrefixRequest {
public:
    SubTemperature() : ScriptSubstitutionPrefixRequest("temp") {}
    virtual ScriptVariable Handle(ScriptVector &params) const  {
        // params[0] is the temp. sensor id, the rest is ignored
        int r = onewire_get_temperature(params[0].c_str());
        if(r < -273)
            return "";
        else
            return ScriptNumber(r);
    }
};

class SubStateOf : public ScriptSubstitutionPrefixRequest {
public:
    SubStateOf() : ScriptSubstitutionPrefixRequest("stateof") {}
    virtual ScriptVariable Handle(ScriptVector &params) const  {
        // params[0] is the automata id ("" for the main)
        return get_current_state_of(params[0].c_str());
    }
};

FFazendaSubstitution::FFazendaSubstitution()
{
    AddVar(new SubCurrentState);
    AddVar(new SubSystemName);
    AddVar(new SubDevReport);
    AddVar(new SubOWReport);
    AddVar(new SubFullReport);
    AddVar(new SubDateReport);
    AddVar(new SubRunVariables);
    AddVar(new SubDevLines);
    AddVar(new SubTemperature);
    AddVar(new SubStateOf);
}

ScriptVariable FFazendaSubstitution::
operator()(const ScriptVariable &src) const
{
    return Substitute(src);
}

ScriptVariable FFazendaSubstitution::
operator()(const ScriptVariable &src, const ScriptVector &argv,
           int idx, int count) const
{
    return Substitute(src, argv, idx, count);
}

ScriptVariable FFazendaSubstitution::
operator()(const ScriptVariable &src, const ScriptVector &argv,
                                      const ScriptVector &dict) const
{
    DictionaryVar dv(argv, dict);
    return Substitute(src, &dv);
}

FFazendaSubstitution FFSubstitution;
