//  FFazenda vers. 1.65                     
//  Copyright (c) Unicontrollers, Ltd., 2013--2016.  All rights reserved.
//  Written by Andrey V. Stolyarov.
// 
//  This is free software.  Permission is granted to everyone to use, copy 
//  or modify this software under the terms and conditions of 
//  GNU GENERAL PUBLIC LICENSE, v.3.0
//  as published by Free Software Foundation (see the file gpl-3.0.txt)   
//                                                                         
//  Please visit http://www.unicontrollers/soft/ffazenda to get a fresh copy
//  This code is provided strictly and exclusively on the "AS IS" basis. 
//  THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED NOR
//  WHATEVER YOU CAN, OR CAN'T, IMAGINE.  IF THIS IS IMPOSSIBLE DUE TO LOCAL
//  LAW REGULATIONS, THEN THE LICENCE IS AUTOMATICALLY REVOKED FOR USERS WHO
//  LIVE IN SUCH LEGAL SYSTEMS.
// 




#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <sue/sue_sel.hpp>
#include <sue/sue_tcps.hpp>
#include <sue/sue_http.hpp>
#include <sue/sue_digs.hpp>

#include <ip4areas/ip4areas.hpp>

#include <scriptpp/scrsubs.hpp>


#include "config.hpp"
#include "flog.hpp"
#include "commands.hpp"
#include "textsubs.hpp"
#include "webserv.hpp"

enum { http_session_timeout = 20 };


////////////////////////////////////////////////////////////

class WebSubVar : public ScriptSubstitution::Var {
public:
    ScriptVariable Sub(const char *name, int len) const;
};

class WebSubs : public FFazendaSubstitution  {
    ScriptVariable cmdresponse;
public:
    WebSubs() : FFazendaSubstitution() {
        AddVar(new ScriptSubstitutionScrVar("cmdresponse", &cmdresponse));
        AddVar(new WebSubVar);
    } 
    void SetCmdResponse(const ScriptVariable &v) { cmdresponse = v; }
};

static WebSubs WebSubstitution;

ScriptVariable WebSubVar::Sub(const char *name, int len) const
{
    if(len < 6 || strncmp("html:", name, 5) != 0) return ScriptVariableInv();
    char *namestr = new char[len+1 - 5];
    memcpy(namestr, name+5, len);
    namestr[len-5] = 0;
    const char *cont = config_get_webserver_content(namestr);
    delete[] namestr;
    if(!cont) return ScriptVariable("");
    return WebSubstitution(cont);
}



////////////////////////////////////////////////////////////

class HttpServerSession : public SUEHttpServerSession {
    class WebServer *the_server;
    webpage_properties wpr;
public:
    HttpServerSession(int a_fd,  
                      SUEEventSelector *a_selector,
                      WebServer *a_server);
    ~HttpServerSession() {}

    void MessageCompleteHook(SUEMessage &msg);

    void SendThePage(SUEMessage &request);
private:
    void SendCommandPage(SUEMessage &request);
    void SendFile();
    void HandleProtectedRequest(SUEMessage &msg);
};

void HttpServerSession::MessageCompleteHook(SUEMessage &msg)
{
    const char *method = GetMethod();
    if(strcasecmp(method, "GET") != 0 && strcasecmp(method, "POST") != 0) {
        AddResponseHeader("Allow", "GET");
        AbortSession(405, "Method Not Allowed");
        return;
    }
    ScriptVariable localpath = GetLocalpath();
    log_message("webserver: request for [%s] method %s",
                localpath.c_str(), method);
    if(!config_get_webpage_properties(localpath.c_str(), &wpr)) {
        log_message("webserver: not found, aborting session");
        ResponseBody().AddString(
            WebSubstitution(config_get_webserver_p404()).c_str());
        AbortSession(404, "Not found");
    }
    if(wpr.is_public) {
        SendThePage(msg);
    } else {
        HandleProtectedRequest(msg);
    }
}

void HttpServerSession::SendThePage(SUEMessage &request)
{
    if(wpr.type == "page") {
        SetHttp11Response(200, "OK");
        AddResponseHeader("Content-Type", wpr.mime_type.c_str());
        ResponseBody().AddString(WebSubstitution(wpr.content).c_str());
        CommitResponse();
    } else
    if(wpr.type == "file") {
        SendFile();
    } else
    if(wpr.type == "command") {
        SendCommandPage(request);
    } else {
        AbortSession(500, "Server configuraton error");
    }
}

void HttpServerSession::SendCommandPage(SUEMessage &request)
{
    SetHttp11Response(200, "OK");
    AddResponseHeader("Content-Type", wpr.mime_type.c_str());
    if(request.GetBody().Length() > 0) {
        request.GetBody().AddChar(0); // make it a correct string
        SUEStringMapHttpQuery query(request.GetBody().GetBuffer());
        ScriptVariable command;
        const char *cmd = query.Get("command");
        if(cmd) {
            command = cmd;
        } else {
            for(int i = 0; ; i++) {
                ScriptVariable nm(8, "cmd%d", i);
                const char *arg = query.Get(nm.c_str());
                if(!arg)
                    break;
                command += arg;
                command += " ";
            }    
        }
        ScriptVariable resp;
        process_command("trusted", command.c_str(), resp);
        WebSubstitution.SetCmdResponse(resp);
        ResponseBody().AddString(WebSubstitution(wpr.response).c_str());
    } else {
        ResponseBody().AddString(WebSubstitution(wpr.content).c_str());
    }
    CommitResponse();
}

void HttpServerSession::SendFile()
{
    int fd = open(wpr.path.c_str(), O_RDONLY);
    if(fd == -1) {
        int e = errno;
        log_message("webserv: couldn't open %s (to send for page %s)",
                    wpr.path.c_str(), GetLocalpath());
        log_message("webserv: %s: %s", wpr.path.c_str(), strerror(e));
        AbortSession(500, "Server configuration error");
        return;
    }
    char buf[1024];
    int rc;
    while((rc = read(fd, buf, sizeof(buf))) > 0) {
        ResponseBody().AddData(buf, rc);
    }
    if(rc == -1) {
        log_message("webserv: reading %s: %s",
                    wpr.path.c_str(), strerror(errno));
        AbortSession(500, "Server configuration error");
    } else {
        SetHttp11Response(200, "OK");
        AddResponseHeader("Content-Type", wpr.mime_type.c_str());
        CommitResponse();
    }
    close(fd);
}

////////////////////////////////////////////////////////////

class LongWebSession : public SUEWebSession {
    class WebServer *the_server;
public:
    LongWebSession(class WebServer *serv);
    void HandleWebRequest(SUEHttpServerSession *sess,
                          SUEMessage *request,
                          SUEMessage *response,
                          const char *username);
};

////////////////////////////////////////////////////////////

class WebServer : public SUETcpServer, public SUEDigestAuthMachine {
    SUEEventSelector *the_selector;
public:
    WebServer(SUEEventSelector *s, const char *ip, int prt, const char *realm)
        : SUETcpServer(ip, prt), SUEDigestAuthMachine(s, realm, "/ctrl"),
        the_selector(s) {}
    ~WebServer() {}
// from SUETcpServer
    SUETcpServerSession* SpawnSession(int newsessionfd)
        { return new HttpServerSession(newsessionfd, the_selector, this); }
// from SUEDigestAuthMachine
    SUEWebSession* DoMakeWebSession(int idx)
        { return new LongWebSession(this); }
    const char *PasswdForUser(const char *user) const;
};

const char *WebServer::PasswdForUser(const char *user) const
{
    static ScriptVariable sv = config_get_webserver_password_by_user(user);
    if(sv.IsInvalid())
        return 0;
    return sv.c_str();
}

////////////////////////////////////////////////////////////

HttpServerSession::
HttpServerSession(int a_fd, SUEEventSelector *a_selector, WebServer *a_server)
    : SUEHttpServerSession(a_fd, http_session_timeout, a_selector, a_server),
    the_server(a_server)
{} 

LongWebSession::LongWebSession(class WebServer *serv)
    : SUEWebSession(serv), the_server(serv)
{}

void LongWebSession::HandleWebRequest(SUEHttpServerSession *sess,
    SUEMessage *request, SUEMessage *response, const char *username)
{
    static_cast<HttpServerSession*>(sess)->SendThePage(*request);
}

void HttpServerSession::HandleProtectedRequest(SUEMessage &msg)
{
    log_message("ZZZ in HandleProtected");
    the_server->HandleHttpRequest(this, &msg, &Response());
}


////////////////////////////////////////////////////////////

static WebServer *server = 0;

void webserv_init(class SUEEventSelector *selector)
{
    if(server) { delete server; server = 0; }
    const char *ip = config_get_webserver_ip();
    if(!ip) {
        log_message("No webserver ip addr configured, web iface disabled");
        return;
    }
    int port = config_get_webserver_port();
    const char *realm = config_get_webserver_realm();
    server = new WebServer(selector, ip, port, realm);
    server->Up(selector);
}
