// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




/*
  This program is the simplest possible Web (http) server. 
  It listens for TCP connections at the port 80 and responds
  with the same page to ANY of the requests.   
  Unlimited number of clients is allowed (in fact,
  the number is limited by open descriptors limits).

  The timeout is set to 20 sec.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sue/sue_sel.hpp>
#include <sue/sue_tcps.hpp>
#include <sue/sue_822.hpp>
#include <sue/sue_http.hpp>

// #define DEBUG_PRINT    // comment this out to disable debug print

const int the_server_port = 80;
const int the_server_timeout = 2000;
const char *response_page =
    "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">"
    "<html><head><title>Aug. 01, 2013: Internet Strike!</title>"
    "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=koi8-r\">"
    "</head>"
    "<body bgcolor=\"black\" text=\"white\">"
    "<center>"
    "<h1>1  2013</h1>"
    "<h1> -</h1>"
    "<font size=\"+2\">"
    "      ,  "
    "ݣ  &mdash;  .<br/>"
    "    .     ?"
    "</font></center></body></html>";



class WebServerSession : public SUEHttpServerSession {
public:
    WebServerSession(int a_fd, int a_timeout, 
                      SUEEventSelector *a_selector,
                      class WebServer *a_server);

    ~WebServerSession();

    void MessageCompleteHook(SUEMessage &msg);

    void SendThePage();
};

class WebServer : public SUETcpServer {
    int timeout;
public:
    WebServer(int a_port, int a_timeout);
    ~WebServer();

    virtual SUETcpServerSession* SpawnSession(int newsessionfd);
};



WebServerSession::WebServerSession(int a_fd, int a_timeout, 
	      SUEEventSelector *a_selector,
	      WebServer *a_server)
    : SUEHttpServerSession(a_fd, a_timeout, a_selector, a_server)
{
} 

WebServerSession::~WebServerSession() 
{ 
#ifdef DEBUG_PRINT
    printf("** session dies **\n");
#endif
}

void WebServerSession::MessageCompleteHook(SUEMessage &msg)
{
    const char *method = GetMethod();
    if(strcasecmp(method, "GET") != 0 && strcasecmp(method, "POST") != 0) {
        AddResponseHeader("Allow", "GET");
        AbortSession(405, "Method Not Allowed");
        return;
    }
    SendThePage();
}

void WebServerSession::SendThePage()
{
    SetHttp11Response(200, "OK");
    AddResponseHeader("Content-Type", "text/html");
    ResponseBody().AddString(response_page);
    CommitResponse();
}

////////////////////////////////////////////////////////////////////////

WebServer::WebServer(int a_port, int a_timeout)
    : SUETcpServer("0.0.0.0", a_port), timeout(a_timeout)
{}

WebServer::~WebServer()
{} 

SUETcpServerSession* WebServer::SpawnSession(int newsessionfd)
{
    return new WebServerSession(newsessionfd, timeout, selector, this);
}

int main(int argc, char **argv)
{
    SUEEventSelector selector;
    WebServer serv(the_server_port, the_server_timeout);
    if(serv.Up(&selector)) { 
        printf("[web server] Listening port %d\n", the_server_port);
    } else {
        printf("Failed to bring the server up, exiting...\n");
        exit(1);
    }
    selector.Go();
    return 0;
}
