// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




/*
  This program is a very simple Web (http) server. 
  It listens for TCP connections at the port 8080.   
  Unlimited number of clients is allowed (in fact,
  the number is limited by open descriptors limits).

  The timeout is set to 20 sec.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sue/sue_sel.hpp>
#include <sue/sue_tcps.hpp>
#include <sue/sue_822.hpp>
#include <sue/sue_http.hpp>
#include <sue/sue_digs.hpp>

#define DEBUG_PRINT    // comment this out to disable debug print

const int the_server_port = 8080;
const int the_server_timeout = 2000;



class WebServerSession : public SUEHttpServerSession {
    static class DigestMachine *digma;
public:
    WebServerSession(int a_fd, int a_timeout, 
                      SUEEventSelector *a_selector,
                      class WebServer *a_server);

    ~WebServerSession();

    void MessageCompleteHook(SUEMessage &msg);
    bool HeaderHook(const char *name, const char *value);

    void SendGenericPage(SUEMessage &msg);
    void SendSimpleForm();
    void HandleTheForm(SUEMessage &msg);
    void HandleProtected(SUEMessage &msg);
};

DigestMachine *WebServerSession::digma = 0;

class WebServer : public SUETcpServer {
    int timeout;
public:
    WebServer(int a_port, int a_timeout);
    ~WebServer();

      // this is the only one we really MUST override in a child of
      // the SUETcpServer class. It must create another WebServerSession
      // object.
    virtual SUETcpServerSession* SpawnSession(int newsessionfd);
};

class LongWebSession : public SUEWebSession {
public:
    LongWebSession(SUEWebSessionCollection *c) : SUEWebSession(c) {}
    virtual void HandleWebRequest(SUEHttpServerSession *sess,
                                  SUEMessage *request,
                                  SUEMessage *response,
                                  const char *username);
};

class DigestMachine : public SUEDigestAuthMachine {
public:
    DigestMachine(SUEEventSelector *sel)
        : SUEDigestAuthMachine(sel, "abc", "/prot") {}
    virtual const char *PasswdForUser(const char *user) const;
    virtual SUEWebSession* DoMakeWebSession(int idx)
        { return new LongWebSession(this); }
};

static const char* passwords[] = {
    "sue", "TheLibrary", 
    "john", "ABCDEFGH",
    "george", "my cat is damn brown",
    0, 0
};

const char *DigestMachine::PasswdForUser(const char *user) const
{
    int i;
    for(i = 0; passwords[i]; i += 2)
        if(0 == strcmp(passwords[i], user))
            return passwords[i+1];
    return 0;
}

WebServerSession::WebServerSession(int a_fd, int a_timeout, 
	      SUEEventSelector *a_selector,
	      WebServer *a_server)
    : SUEHttpServerSession(a_fd, a_timeout, a_selector, a_server)
{
    if(!digma) digma = new DigestMachine(a_selector);
} 

WebServerSession::~WebServerSession() 
{ 
#ifdef DEBUG_PRINT
    printf("** session dies **\n");
#endif
}

void WebServerSession::MessageCompleteHook(SUEMessage &msg)
{
    const char *method = GetMethod();
    if(strcasecmp(method, "GET") != 0 && strcasecmp(method, "POST") != 0) {
        AddResponseHeader("Allow", "GET");
        AbortSession(405, "Method Not Allowed");
        return;
    }
    if(strcmp(GetLocalpath(), "/form") == 0)
        SendSimpleForm();
    else
    if(strncmp(GetLocalpath(), "/handler", 8) == 0)
        HandleTheForm(msg);
    else
    if(strcmp(GetLocalpath(), "/prot") == 0 ||
        strncmp(GetLocalpath(), "/prot/", 6) == 0
    )
        HandleProtected(msg);
    else
        SendGenericPage(msg);
}

void WebServerSession::SendGenericPage(SUEMessage &msg)
{
    SetHttp11Response(200, "OK");
    AddResponseHeader("Content-Type", "text/html");
    AddResponseHeader("Set-Cookie", "mycookie=myvalue");
    ResponseBody().AddString("<html><body><center>IT WORKS</center>\n");
    ResponseBody().AddString("<ul><li>Method: ");
    ResponseBody().AddString(GetMethod());
    ResponseBody().AddString("</li><li>Host: ");
    ResponseBody().AddString(GetHost());
    ResponseBody().AddString("</li><li>Localpath: ");
    ResponseBody().AddString(GetLocalpath());
    ResponseBody().AddString("</li><li>Query: ");
    ResponseBody().AddString(GetQuery() ? GetQuery() : "");
    ResponseBody().AddString("</li><li>Original URI: ");
    ResponseBody().AddString(GetUri());
    int len = msg.GetBody().Length();
    if(len > 0) {
        ResponseBody().AddString("</li><li>Body Length: ");
        char b[16];
        sprintf(b, "%d", len);
        ResponseBody().AddString(b);
        ResponseBody().AddString("</li><li>Body:\n<pre>\n");
        ResponseBody().AddData(msg.GetBody().GetBuffer(), len);
        ResponseBody().AddString("\n<pre>\n");
    }
    ResponseBody().AddString("</li></ul>\n"
                             "<a href=\"/form\">Go to the form</a><br>\n"
                             "<a href=\"/prot\">Go to protected area</a>\n"
                             "</body></html>\n");
    CommitResponse();
}

void WebServerSession::SendSimpleForm()
{
    SetHttp11Response(200, "OK");
    AddResponseHeader("Content-Type", "text/html");
    AddResponseHeader("Set-Cookie", "mycookie=myvalue");
    ResponseBody().AddString("<html><body>\n");
    ResponseBody().AddString("<h1>This is a POST-form</h1>\n"
        "<form name=\"form1\" action=\"/handler\" method=\"POST\">\n"
        "<input type=\"text\" name=\"user\">"
        "<input type=\"text\" name=\"comment\">"
        "<input type=\"submit\" value=\"Submit it!\">"
        "</form>\n"
    );
    ResponseBody().AddString("<h1>This is a GET-form</h1>\n"
        "<form name=\"form2\" action=\"/handler\" method=\"GET\">\n"
        "<input type=\"text\" name=\"user\">"
        "<input type=\"text\" name=\"comment\">"
        "<input type=\"submit\" value=\"Submit it!\">"
        "</form>\n"
    );
    ResponseBody().AddString("</body></html>\n");
    CommitResponse();
}

void WebServerSession::HandleTheForm(SUEMessage &msg)
{
#ifdef DEBUG_PRINT
    printf("*-<<<");
    for(int i = 0; i<msg.GetBody().Length(); i++)
        printf("%c", msg.GetBody().GetBuffer()[i]);
    printf(">>>-*\n");
#endif
    SendGenericPage(msg);
}

bool WebServerSession::HeaderHook(const char *name, const char *value)
{
#ifdef DEBUG_PRINT
    printf("*[%s]:[%s]*\n", name, value);
#endif
    if(SUEHttpServerSession::HeaderHook(name, value))
        return true;
    return false;
}

void LongWebSession::HandleWebRequest(SUEHttpServerSession *sess,
        SUEMessage *request, SUEMessage *response, const char *username)
{
    sess->SetHttp11Response(200, "OK");
    sess->AddResponseHeader("Content-Type", "text/html");
    response->GetBody().AddString("<html><body><center>YOU'RE IN</center>\n");
    response->GetBody().AddString("</body></html>");
    sess->CommitResponse();
}

void WebServerSession::HandleProtected(SUEMessage &msg)
{
    digma->HandleHttpRequest(this, &msg, &Response());
}


////////////////////////////////////////////////////////////////////////

WebServer::WebServer(int a_port, int a_timeout)
    : SUETcpServer("0.0.0.0", a_port), timeout(a_timeout)
{}

WebServer::~WebServer()
{} 

SUETcpServerSession* WebServer::SpawnSession(int newsessionfd)
{
    return new WebServerSession(newsessionfd, timeout, selector, this);
}

int main(int argc, char **argv)
{
    SUEEventSelector selector;
    WebServer serv(the_server_port, the_server_timeout);
    if(serv.Up(&selector)) { 
        printf("[web server] Listening port %d\n", the_server_port);
    } else {
        printf("Failed to bring the server up, exiting...\n");
        exit(1);
    }
    selector.Go();
    return 0;
}
