// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#include <string.h>
#include <stdio.h>  // for sprintf
#include <stdlib.h> // for strtol
#include <time.h>
#include "sue_http.hpp"

#ifndef SUE_HTTP_DATE_FORMAT
#define SUE_HTTP_DATE_FORMAT "%a, %d %b %Y %H:%M:%S GMT"
#endif


SUEHttpServerSession::SUEHttpServerSession(int fd, int timeout, 
                     SUEEventSelector *selector,
                     SUETcpServer *server)
    : SUETcpServerSession(fd, timeout, selector, server),
    request(), response(), state(start), assume_http_1_1(false),
    recv_headers_count(0), recv_body_count(0),
    expected_body_length(-1), persistent_allowed(false),
    method(0), uri(0), host(0), localpath(0), query(0)
{
    inputresetstimeout = true;
    outputresetstimeout = true;
}

SUEHttpServerSession::~SUEHttpServerSession()
{
    if(method)
        delete[] method;
    if(uri)
        delete[] uri;
    if(host)
        delete[] host;
    if(localpath)
        delete[] localpath;
    // please note there's INTENTIONALLY no deletion for query
}


void SUEHttpServerSession::DropIncomingMessage()
{
    request.DropAll();
}

void SUEHttpServerSession::SetHttp11Response(int code, const char *comment)
{
    char *buf = new char[strlen(comment) + sizeof("HTTP/1.1 000  ")];
    code %= 1000;
    sprintf(buf, "HTTP/1.1 %3d %s", code, comment);
    response.SetHeadingLine(buf);
    delete[] buf;
}

void SUEHttpServerSession::
AddResponseHeader(const char *name, const char *value)
{
    response.AddHeader(name, value);
}

void SUEHttpServerSession::AddResponseHeaderDate()
{
    char timebuf[32];
    time_t now = time(0);
    strftime(timebuf, sizeof(timebuf), SUE_HTTP_DATE_FORMAT, gmtime(&now));
    AddResponseHeader("Date", timebuf);
}

void SUEHttpServerSession::AddResponseHeaderServer()
{
// XXX this is to be changed!!!
    AddResponseHeader("Server", "Another ugly server created with SUE");
}

void SUEHttpServerSession::AddResponseHeaderContentlen()
{
    char buf[16];  // more than enough for 32-bit decimal
    sprintf(buf, "%d", response.GetBody().Length());
    AddResponseHeader("Content-Length", buf);
}

void SUEHttpServerSession::CommitResponse()
{
    response.Export(outputbuffer);
    FinishSingleRequest();
}

void SUEHttpServerSession::SendLongResponse()
{
    response.Export(outputbuffer);
    state = longresponse;
}

void SUEHttpServerSession::AbortSession(int code, const char *comment,
                  const char *longcomment, bool http11)
{
    if(state == responded_closing)
        return;
    if(!comment)
        comment = "Error";
//    response.GetBody().DropAll();
    if(response.GetBody().Length() == 0) {
        if(!longcomment)
            longcomment = comment;
        response.GetBody().AddString(
            "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n"
            "<HTML><HEAD><TITLE>\n");
        char buf[8];
        code %= 1000;
        sprintf(buf, "%3d ", code);
        response.GetBody().AddString(buf);
        response.GetBody().AddString(comment);
        response.GetBody().AddString("</TITLE></HEAD><BODY><H1>");
        response.GetBody().AddString(comment);
        response.GetBody().AddString("</H1><P>");
        response.GetBody().AddString(longcomment);
        response.GetBody().AddString("</P></BODY></HTML>\n");
    }
    if(http11) {
        SetHttp11Response(code, comment);
        AddResponseHeaderDate();
        AddResponseHeaderServer();
        AddResponseHeader("Content-Type", "text/html; charset=iso-8859-1");
        AddResponseHeaderContentlen();
    }
    persistent_allowed = false;
    CommitResponse();
}

void SUEHttpServerSession::BrokenProtocol(bool http11)
{
    AbortSession(400, "Bad Request", "I don't understand you", http11);
}

static bool is_char_spc(int ch)
{
    return ch == ' ' || ch == '\t' || ch == '\r';
}

void SUEHttpServerSession::HandleNewInput()
{
    bool body_hook = false;
    for(int i = 0; i < inputbuffer.Length(); i++) {
        switch(state) {
            case incoming_complete:
            case longresponse:
            case responded_closing:
                inputbuffer.DropData(i); // in case we got something...
                DisableInput();
                goto process_hooks;
            default:
                ;
        }
        char ch = inputbuffer[i];
        const char *req;
        request.FeedChar(ch);
        switch(state) {
            case start:
                if(ch != '\n' && !is_char_spc(ch))
                    state = getting_request;
                break;
            case getting_request:
                if((req = request.GetHeadingLine()) != 0) {
                    AnalyseRequestLine(req);
                    state = in_header;
                }
                break;
            case in_header:
                if(recv_headers_count < request.GetHeaderCount()) {
                    AnalyseHeaderLine(request.GetLastHeaderName(),
                                      request.GetLastHeaderBody());
                    recv_headers_count = request.GetHeaderCount();
                }
                if(request.HeaderEnded()) {
                    if(expected_body_length > 0) {
                        state = in_body;
                        recv_body_count = 0;
                    } else {
                        state = incoming_complete;
                        MessageCompleteHook(request);
                    }
                }
                break;
            case in_body:
                recv_body_count++;
                body_hook = true;
                if(expected_body_length >= 0 &&
                    recv_body_count >= expected_body_length)
                {
                    state = incoming_complete;
                    BodyProcessHook(request.GetBody());
                    BodyCompleteHook(request.GetBody());
                    MessageCompleteHook(request);
                    body_hook = false; // already done
                }
                break;

            case incoming_complete:
            case longresponse:
            case responded_closing:
            default:
                // actually, this should never ever happen
                ;
        }
    }
    inputbuffer.DropAll();
process_hooks:
    if(body_hook)
        BodyProcessHook(request.GetBody());
    CheckSizes();
}

void SUEHttpServerSession::HandleSessionTimeout()
{
    GracefulShutdown();
}

void SUEHttpServerSession::OutputEmptyHook()
{
    if(state != longresponse)
        return;
    if(!LongResponsePortion(outputbuffer))
        FinishSingleRequest();
}

void SUEHttpServerSession::FinishSingleRequest()
{
    if(persistent_allowed) {
        request.DropAll();
        response.DropAll();
        state = start;
        recv_headers_count = 0;
        recv_body_count = 0;
        expected_body_length = -1;
        DisableInput(false);
    } else {
        state = responded_closing;
        GracefulShutdown();
    }
}

void SUEHttpServerSession::UrlDecode(char *str)
{
    char *r=str;
    char *w=str;
    char tmp[3]={0,0,0};
    while (*r){
        switch(*r) {
        case '%':
            if(r[1] && r[2]) {
                tmp[0]=r[1];
                tmp[1]=r[2];
                *w=strtol(tmp, 0, 16);
                r += 2; // two extra chars consumed
            } else {
                *w = '%'; // actually the behaviour is undefined...
                          // so here is our version of "undefined" :)
            }
            break;
        case '+':
            *w=' ';
            break;
        default:
            *w=*r;
        }
        r++;
        w++;
    }
    *w='\0';
}


void SUEHttpServerSession::AnalyseRequestLine(const char *req)
{
    while(*req && is_char_spc(*req))
        req++;
    int method_len = 0;
    while(req[method_len] && !is_char_spc(req[method_len]))
        method_len++;
    if(method_len < 1) {
        BrokenProtocol(false);
        return;
    }

    if(method) delete[] method;
    if(uri) delete[] uri;
    if(host) delete[] host;
    if(localpath) delete[] localpath;
    // please note there's INTENTIONALLY no deletion for query
    method = uri = host = localpath = query = 0;

    method = new char[method_len+1];
    int i;

    for(i = 0; i < method_len; i++)
        method[i] = req[i];
    method[method_len] = 0;
    req += method_len;
    while(*req && is_char_spc(*req))
        req++;
    const char *p = req;
    while(*p) p++;  // past the string
    while(p > req && is_char_spc(*(p-1))) p--;
        // now at the first trailing space, if any, or at the TZero
    const char *first_trailing_space = p;
    while(p > req && !is_char_spc(*(p-1))) p--;
        // now at the (presumably) beginning of protocol token


    int proto_len = first_trailing_space - p;
    char *proto = new char[proto_len + 1]; 
    for(i = 0; i < proto_len; i++)
        proto[i] = p[i];
    proto[proto_len] = 0;

    if(p > req) {
        assume_http_1_1 = (0 == strcasecmp(proto, "HTTP/1.1"));
        while(p > req && is_char_spc(*(p-1))) p--;

        uri = new char[p - req + 1];
        for(i = 0; i < p - req; i++)
            uri[i] = req[i];
        uri[p - req] = 0;
    } else {
        // no protocol token
        // seems like older protocol
        uri = proto;
        proto = 0;
        assume_http_1_1 = false;
    }
    if(proto) delete[] proto;   // it's local
    RequestLineHook(method, uri, assume_http_1_1);
    if(uri[0] == '*' && uri[1] == 0)
        return;
    if(uri[0] == '/') {
        localpath = new char[strlen(uri)+1];
        strcpy(localpath, uri);
    } else {  // full URL
        char *h = strstr(uri, "://");
        if(!h) {
            AbortSession();
            return;
        }
        h += 3;
        char *u = strchr(h, '/');
        if(u) {
            host = new char[u - h];
            strncpy(host, h, u - h);
            host[u - h - 1] = 0;
            u++;
            localpath = new char[strlen(u)+1];
            strcpy(localpath, u);
        } else {
            // the client is buggy, sent us no localpart, but we'll serve
            localpath = new char[2];
            localpath[0] = '/'; localpath[1] = 0;
            host = new char[strlen(h)+1];
            strcpy(host, h);
        }
    }
    // well may be there's also query
    char *qq = strchr(localpath, '?');
    if(qq) {
        query = qq + 1;
        *qq = 0; // so that localpath and query are separated
    }
    UrlDecode(localpath);
}

void SUEHttpServerSession::
AnalyseHeaderLine(const char *name, const char *body)
{
    if(strcmp(name, "Content-Length") == 0) {
        long nn;
        char *err;
        nn = strtol(body, &err, 10);
        if(err == body || *err != '\0') {
            AbortSession();
            return;
        }
        expected_body_length = nn;
        if(!CheckSizes())
            return;
    }
    if(strcmp(name, "Host") == 0) {
        if(!host) {
            host = new char[strlen(body)+1];
            strcpy(host, body);
        }
    }
    HeaderHook(name, body);
}

bool SUEHttpServerSession::HeaderHook(const char *name, const char *value)
{
    if(strcmp(name, "Expect") == 0) {
        AbortSession(417, "Expectation failed");
        return true;
    }
    return false;
}

bool SUEHttpServerSession::CheckSizes()
{
    if(AreSizesAcceptable(request.HeaderSize(), expected_body_length))
        return true;
    AbortSession(413, "Request Entity Too Large");
    return false;
}



SUEStringMapHttpQuery::SUEStringMapHttpQuery(const char *from)
{
    enum states { inname, invalue };
    states state = inname;
    const char *mark = from;
    char *name = 0, *value = 0;
    const char *p;
    for(p = from; *p; p++) {
        switch(state) {
        case inname:
            if(*p == '=') {
                int len = p - mark;
                name = new char[len+1];
                memcpy(name, mark, len);
                name[len] = 0;
                state = invalue;
                mark = p+1;
            }
            break;
        case invalue:
            if(*p == '&') {
                int len = p - mark;
                value = new char[len+1];
                memcpy(value, mark, len);
                value[len] = 0;
                state = inname;
                SUEHttpServerSession::UrlDecode(name);
                SUEHttpServerSession::UrlDecode(value);
                Add(name, value, true);
                name = value = 0;
                mark = p+1;
            }
            break;
        }
    }
    if(name && state == invalue) {
        int len = p - mark;
        value = new char[len+1];
        strcpy(value, mark);
        SUEHttpServerSession::UrlDecode(name);
        SUEHttpServerSession::UrlDecode(value);
        Add(name, value, true);
        name = value = 0;
    }
    if(name)
        delete[] name;
    if(value)
        delete[] value;

}
