// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#ifndef SENTRY_SUE_HTTP_HPP
#define SENTRY_SUE_HTTP_HPP

#include "sue_tcps.hpp"
#include "sue_822.hpp"

class SUEHttpServerSession : public SUETcpServerSession  {
    SUEMessageIncoming request;
    SUEMessage response;
    enum states {
        start, getting_request, in_header, in_body, incoming_complete,
        longresponse, responded_closing
    } state;
    bool assume_http_1_1;
    int recv_headers_count;
    int recv_body_count;
    int expected_body_length;
    
    bool persistent_allowed;

    char *method, *uri, *host, *localpath, *query;
        // pleae note query must either point into the localpath-pointed
        // memory, or be NULL, so it must never be delete[]'d
public:
    SUEHttpServerSession(int a_fd, int a_timeout, 
                         SUEEventSelector *a_selector,
                         SUETcpServer *server);

    virtual void RequestLineHook(const char *method,
                                 const char *uri,
                                 bool is_1_1) {}
    virtual bool HeaderHook(const char *name, const char *value);
    virtual void HeaderCompleteHook(SUEMessage &msg) {}
    virtual void BodyProcessHook(SUEBuffer &body) {}
    virtual void BodyCompleteHook(SUEBuffer &body) {}
    virtual void MessageCompleteHook(SUEMessage &msg) {}

    const char *GetMethod() const { return method; }
    const char *GetUri() const { return uri; }
    const char *GetHost() const { return host; }
    const char *GetLocalpath() const { return localpath; }
    const char *GetQuery() const { return query; }

    virtual bool AreSizesAcceptable(int hdr, int bdy) const
        { return hdr <= 0x4000; }

    void AllowPersistent(bool a = true) { persistent_allowed = a; }
    void DropIncomingMessage();
    void SetHttp11Response(int code, const char *comment);
    void AddResponseHeader(const char *name, const char *value);
    void AddResponseHeaderDate();
    void AddResponseHeaderServer();
    void AddResponseHeaderContentlen();
    SUEBuffer& ResponseBody() { return response.GetBody(); }
    SUEMessage& Response() { return response; }
    void CommitResponse();
    void SendLongResponse();
    virtual bool LongResponsePortion(SUEBuffer &outbuf) { return false; }
    void AbortSession(int code = 400, const char *comment = 0,
                      const char *long_comment = 0, bool http11 = true);
    virtual void BrokenProtocol(bool http11 = true);

        //! Helper function that decodes urlencoded data 'in place'
    static void UrlDecode(char *str);

protected:
        //! Destructor
        /*! \note Destructor made protected because only SUETcpServer is
            allowed to destroy its sessions 
         */
    ~SUEHttpServerSession();

private:
    virtual void HandleNewInput();
    virtual void HandleSessionTimeout();
    virtual void OutputEmptyHook();

    void FinishSingleRequest();
    void AnalyseRequestLine(const char *req);
    void AnalyseHeaderLine(const char *name, const char *body);
    bool CheckSizes();
};


class SUEStringMapHttpQuery : public SUEStringMap {
public:
    SUEStringMapHttpQuery(const char *from);
};


#endif
