// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#include <string.h>
#include <unistd.h>
#include "sue_sess.hpp"






// SUEGenericDuplexSession

void SUEGenericDuplexSession::TimeoutHandle()
{
    HandleSessionTimeout();
}

void SUEGenericDuplexSession::FdHandle(bool a_r, bool a_w, bool /*a_ex*/)
{
    if(a_r && inputenabled) {
        char buf[1024];
        int rb = read(fd, buf, sizeof(buf));
        if(rb>0) {
            if(inputresetstimeout)
                ResetTimeout();
            inputbuffer.AddData(buf, rb);
            HandleNewInput();
        } else if(rb==0) {
            HandleRemoteClosing();
            return;
        } else {
            HandleReadError();
        }
    }
    if(a_w) {
        int len = outputbuffer.Length();
        if(len) {
            int wb = write(fd, outputbuffer.GetBuffer(), len);
            if(wb>0) { 
                if(outputresetstimeout)
                    ResetTimeout();
                outputbuffer.DropData(wb);
            } else if(wb==-1) {
                HandleWriteError();
            } else {
                 /* write returned zero, heh... We simply ignore this
                    situation as it seems to be impossible anyway
                  */
            }
            if(outputbuffer.Length() == 0)
                 OutputEmptyHook();
            if(scheduleddown && outputbuffer.Length() == 0) 
                 Shutdown();
        }
    }
}

SUEGenericDuplexSession::SUEGenericDuplexSession( 
                            int a_timeout, 
                            const char *a_greeting) 
   : SUEFdHandler(), SUETimeoutHandler()  
{
    the_selector = 0;  
    timeout_sec = a_timeout;
    timeout_usec = 0;
    if(a_greeting) {
        outputbuffer.AddData(a_greeting, strlen(a_greeting));
    }
    scheduleddown = false;
    inputenabled = true;
    inputresetstimeout = true;
    outputresetstimeout = true;
}

SUEGenericDuplexSession::~SUEGenericDuplexSession()
{
    Shutdown();
}

void SUEGenericDuplexSession::Startup(SUEEventSelector *a_selector, 
                                      int a_fd, const char *a_greeting)
{
    the_selector = a_selector;
    if(a_greeting) {
        outputbuffer.DropAll();
        outputbuffer.AddData(a_greeting, strlen(a_greeting));
    }
    SetFd(a_fd);
    the_selector->RegisterFdHandler(this);
    if(timeout_sec>0 || timeout_usec>0) {
        SetFromNow(timeout_sec, timeout_usec);
        the_selector->RegisterTimeoutHandler(this);
    }
}

void SUEGenericDuplexSession::Shutdown()
{
    if(the_selector) {
        the_selector->RemoveFdHandler(this);
        the_selector->RemoveTimeoutHandler(this);
    }
    the_selector = 0;
    if(fd != -1) {
        close(fd);
        SetFd(-1);
        ShutdownHook();
    }
}

void SUEGenericDuplexSession::GracefulShutdown()
{
    scheduleddown = true;
    if(outputbuffer.Length() == 0) Shutdown();
}

void SUEGenericDuplexSession::ResetTimeout()
{
    the_selector->RemoveTimeoutHandler(this);
    if(timeout_sec>0 || timeout_usec>0) {
        SetFromNow(timeout_sec, timeout_usec);
        the_selector->RegisterTimeoutHandler(this);
    }
}

void SUEGenericDuplexSession::SetTimeout(int sec, int usec)
{
    timeout_sec = sec;
    timeout_usec = usec;
    if(the_selector) ResetTimeout();
}
