// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#include <string.h>
#include "sue_smap.hpp"


SUEStringMap::SUEStringMap()
    : first(0), last(0), item_count(0)
{
}

SUEStringMap::~SUEStringMap()
{
    DropAll();
}

void SUEStringMap::DropAll()
{
    while(first) {
        item *tmp = first;
        first = tmp->next;
        delete tmp;
    }
    first = last = 0;
    item_count = 0;
}

void SUEStringMap::Add(const char *name, const char *body, bool own_them)
{
    item *t = new item;
    if(own_them) {
        t->name = const_cast<char*>(name);
        t->body = const_cast<char*>(body);
    } else {
        t->name = Strdup(name);
        t->body = Strdup(body);
    }
    t->next = 0;
    if(last) {
        last->next = t;
        last = t;
    } else {
        first = last = t;
    }
    item_count++;
}

void SUEStringMap::Set(const char *name, const char *body)
{
    item *t = Find(name);
    if(t) {
        if(t->body)
            delete[] t->body;
        t->body = Strdup(body);
    } else {
        Add(name, body);
    }
}

bool SUEStringMap::Remove(const char *name, bool all)
{
    bool res = false; 
    item **p = &first;
    while(*p) {
        if(0 == strcmp((*p)->name, name)) {
            res = true;
            item *t = *p;
            *p = (*p)->next;
            delete t;
            item_count--;
            if(!all)
                break;
        } else {
            p = &((*p)->next);
        }
    }
    return res;
}

const char *SUEStringMap::Get(const char *name, bool ignorecase) const
{
    item *t = Find(name, ignorecase);
    return t ? t->body : 0;
}

const char *SUEStringMap::GetLastName() const
{
    return last ? last->name : 0;
}

const char *SUEStringMap::GetLastBody() const
{
    return last ? last->body : 0;
}

SUEStringMap::item* SUEStringMap::Find(const char *name, bool igncase) const
{
    item* t = first;
    while(t && 0 != (igncase ? strcasecmp : strcmp)(t->name, name))
        t = t->next;
    return t;
}

char* SUEStringMap::Strdup(const char *str) {
    char *res = new char[strlen(str)+1];
    strcpy(res, str);
    return res;
}
