// +-------------------------------------------------------------------------+
// |              (S)imple (U)nix (E)vents vers. 0.3.01                      |
// | Copyright (c) Andrey Vikt. Stolyarov <crocodil_AT_croco.net> 2003-2013. |
// | ----------------------------------------------------------------------- |
// | This is free software.  Permission is granted to everyone to use, copy  |
// |        or modify this software under the terms and conditions of        |
// |                 GNU LESSER GENERAL PUBLIC LICENSE, v. 2.1               |
// | as published by Free Software Foundation       (see the file LGPL.txt)  |
// |                                                                         |
// | Please visit http://www.croco.net/software/sue for a fresh copy of Sue. |
// | ----------------------------------------------------------------------- |
// |   This code is provided strictly and exclusively on the "AS IS" basis.  |
// | !!! THERE IS NO WARRANTY OF ANY KIND, NEITHER EXPRESSED NOR IMPLIED !!! |
// +-------------------------------------------------------------------------+




#ifndef SENTRY_SUE_SMAP_HPP
#define SENTRY_SUE_SMAP_HPP

//! This class implements a string map, a set of NAME/VALUE pairs
class SUEStringMap {
    struct item {
        char *name;
        char *body;
        item *next;
        ~item() { if(name) delete[] name; if(body) delete[] body; }
    };
    item *first, *last;
    int item_count;

public:
    SUEStringMap();
    ~SUEStringMap();

    void DropAll();

    void Add(const char *name, const char *body, bool own_them = false);
    void Set(const char *name, const char *body);
    bool Remove(const char *name, bool all = false);
    const char *Get(const char *name, bool ignorecase = false) const;
    const char *GetLastName() const;
    const char *GetLastBody() const;
    int GetCount() const { return item_count; }

    friend class Iterator;
    class Iterator {
        const item *current;
    public:
        Iterator(const SUEStringMap *master) : current(master->first) {}
        bool Next() { return (current = current ? current->next : 0); }
        const char *Name() const { return current ? current->name : 0; }
        const char *Body() const { return current ? current->body : 0; }
    };
private:
    item* Find(const char *name, bool ignorecase = false) const;
    static char *Strdup(const char *str);
};

#endif
